"""Temperature indicator definitions."""

from __future__ import annotations

from xclim import indices
from xclim.core import cfchecks
from xclim.core.indicator import (
    Daily,
    Hourly,
    Indicator,
    ResamplingIndicatorWithIndexing,
)
from xclim.core.utils import InputKind

__all__ = [
    "australian_hardiness_zones",
    "biologically_effective_degree_days",
    "chill_portions",
    "chill_units",
    "cold_spell_days",
    "cold_spell_duration_index",
    "cold_spell_frequency",
    "cold_spell_max_length",
    "cold_spell_total_length",
    "consecutive_frost_days",
    "cool_night_index",
    "cooling_degree_days",
    "daily_freezethaw_cycles",
    "daily_temperature_range",
    "daily_temperature_range_variability",
    "degree_days_exceedance_date",
    "extreme_temperature_range",
    "fire_season",
    "first_day_tg_above",
    "first_day_tg_below",
    "first_day_tn_above",
    "first_day_tn_below",
    "first_day_tx_above",
    "first_day_tx_below",
    "freezethaw_spell_frequency",
    "freezethaw_spell_max_length",
    "freezethaw_spell_mean_length",
    "freezing_degree_days",
    "freshet_start",
    "frost_days",
    "frost_free_season_end",
    "frost_free_season_length",
    "frost_free_season_start",
    "frost_free_spell_max_length",
    "frost_season_length",
    "growing_degree_days",
    "growing_season_end",
    "growing_season_length",
    "growing_season_start",
    "heat_spell_frequency",
    "heat_spell_max_length",
    "heat_spell_total_length",
    "heat_wave_frequency",
    "heat_wave_index",
    "heat_wave_max_length",
    "heat_wave_total_length",
    "heating_degree_days",
    "hot_spell_frequency",
    "hot_spell_max_length",
    "hot_spell_max_magnitude",
    "hot_spell_total_length",
    "huglin_index",
    "ice_days",
    "last_spring_frost",
    "late_frost_days",
    "latitude_temperature_index",
    "max_daily_temperature_range",
    "maximum_consecutive_frost_free_days",
    "maximum_consecutive_warm_days",
    "tg10p",
    "tg90p",
    "tg_days_above",
    "tg_days_below",
    "tg_max",
    "tg_mean",
    "tg_min",
    "thawing_degree_days",
    "tn10p",
    "tn90p",
    "tn_days_above",
    "tn_days_below",
    "tn_max",
    "tn_mean",
    "tn_min",
    "tropical_nights",
    "tx10p",
    "tx90p",
    "tx_days_above",
    "tx_days_below",
    "tx_max",
    "tx_mean",
    "tx_min",
    "tx_tn_days_above",
    "usda_hardiness_zones",
    "warm_spell_duration_index",
]


# We need to declare the class here so that the `atmos` realm is automatically detected.
class Temp(Daily):
    """Indicators involving daily temperature."""

    keywords = "temperature"


class TempHourly(Hourly):
    """Temperature indicators involving hourly temperature."""

    keywords = "temperature"


class TempWithIndexing(ResamplingIndicatorWithIndexing):
    """Indicators involving daily temperature and adding an indexing possibility."""

    src_freq = "D"
    keywords = "temperature"


class TempHourlyWithIndexing(ResamplingIndicatorWithIndexing):
    """Indicators involving hourly temperature and adding an indexing possibility."""

    src_freq = "h"
    keywords = "temperature"


tn_days_above = TempWithIndexing(
    title="Number of days with minimum temperature above a given threshold",
    identifier="tn_days_above",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="The number of days with minimum temperature above {thresh}",
    description="{freq} number of days where daily minimum temperature exceeds {thresh}.",
    abstract="The number of days with minimum temperature above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tn_days_above,
    parameters={"op": {"default": ">"}},
)

tn_days_below = TempWithIndexing(
    title="Number of days with minimum temperature below a given threshold",
    identifier="tn_days_below",
    units="days",
    standard_name="number_of_days_with_air_temperature_below_threshold",
    long_name="The number of days with minimum temperature below {thresh}",
    description="{freq} number of days where daily minimum temperature is below {thresh}.",
    abstract="The number of days with minimum temperature below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tn_days_below,
    parameters={"op": {"default": "<"}},
)

tg_days_above = TempWithIndexing(
    title="Number of days with mean temperature above a given threshold",
    identifier="tg_days_above",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="The number of days with mean temperature above {thresh}",
    description="{freq} number of days where daily mean temperature exceeds {thresh}.",
    abstract="The number of days with mean temperature above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tg_days_above,
    parameters={"op": {"default": ">"}},
)

tg_days_below = TempWithIndexing(
    title="Number of days with mean temperature below a given threshold",
    identifier="tg_days_below",
    units="days",
    standard_name="number_of_days_with_air_temperature_below_threshold",
    long_name="The number of days with mean temperature below {thresh}",
    description="{freq} number of days where daily mean temperature is below {thresh}.",
    abstract="The number of days with mean temperature below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tg_days_below,
    parameters={"op": {"default": "<"}},
)

tx_days_above = TempWithIndexing(
    title="Number of days with maximum temperature above a given threshold",
    identifier="tx_days_above",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="The number of days with maximum temperature above {thresh}",
    description="{freq} number of days where daily maximum temperature exceeds {thresh}.",
    abstract="The number of days with maximum temperature above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tx_days_above,
    parameters={"op": {"default": ">"}},
)

tx_days_below = TempWithIndexing(
    title="Number of days with maximum temperature below a given threshold",
    identifier="tx_days_below",
    units="days",
    standard_name="number_of_days_with_air_temperature_below_threshold",
    long_name="The number of days with maximum temperature below {thresh}",
    description="{freq} number of days where daily max temperature is below {thresh}.",
    abstract="The number of days with maximum temperature below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tx_days_below,
    parameters={"op": {"default": "<"}},
)

tx_tn_days_above = TempWithIndexing(
    title="Number of days with daily minimum and maximum temperatures exceeding thresholds",
    identifier="tx_tn_days_above",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="Number of days with daily minimum above {thresh_tasmin} "
    "and daily maximum temperatures above {thresh_tasmax}",
    description="{freq} number of days where daily maximum temperature exceeds {thresh_tasmax} and minimum temperature "
    "exceeds {thresh_tasmin}.",
    abstract="Number of days with daily maximum and minimum temperatures above given thresholds.",
    cell_methods="",
    compute=indices.tx_tn_days_above,
)


heat_wave_frequency = Temp(
    title="Heat wave frequency",
    identifier="heat_wave_frequency",
    units="",
    long_name="Total number of series of at least {window} consecutive days with daily minimum temperature above "
    "{thresh_tasmin} and daily maximum temperature above {thresh_tasmax}",
    description="{freq} number of heat wave events within a given period. A heat wave occurs when daily minimum and "
    "maximum temperatures exceed {thresh_tasmin} and {thresh_tasmax}, respectively, over at least {window} days.",
    abstract="Number of heat waves. A heat wave occurs when daily minimum and maximum temperatures exceed given "
    "thresholds for a number of days.",
    cell_methods="",
    keywords="health,",
    compute=indices.heat_wave_frequency,
)

hot_spell_max_magnitude = Temp(
    title="Hot spell maximum magnitude",
    identifier="hot_spell_max_magnitude",
    units="K d",
    long_name="Maximum cumulative difference between daily maximum temperature and {thresh} for days within a heat wave. "
    "A heat wave is defined as a series of at least {window} consecutive days with daily maximum temperature above {thresh}.",
    description="Magnitude of the most intensive heat wave per {freq}. The magnitude is the cumulative exceedance of daily "
    "maximum temperature over {thresh}. A heat wave is defined as a series of at least {window} consecutive days with daily "
    "maximum temperature above {thresh}",
    abstract="Magnitude of the most intensive heat wave per {freq}. A heat wave occurs when daily maximum "
    "temperatures exceed given thresholds for a number of days.",
    cell_methods="",
    compute=indices.hot_spell_max_magnitude,
)

heat_wave_max_length = Temp(
    title="Heat wave maximum length",
    identifier="heat_wave_max_length",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Longest series of at least {window} consecutive days with daily minimum temperature above "
    "{thresh_tasmin} and daily maximum temperature above {thresh_tasmax}",
    description="{freq} maximum length of heat wave events occurring within a given period. "
    "A heat wave occurs when daily minimum and maximum temperatures exceed {thresh_tasmin} and {thresh_tasmax}, "
    "respectively, over at least {window} days.",
    abstract="Maximal duration of heat waves. A heat wave occurs when daily minimum and maximum temperatures exceed "
    "given thresholds for a number of days.",
    cell_methods="",
    keywords="health,",
    compute=indices.heat_wave_max_length,
)

heat_wave_total_length = Temp(
    title="Heat wave total length",
    identifier="heat_wave_total_length",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Total length of events of at least {window} consecutive days with daily minimum temperature above "
    "{thresh_tasmin} and daily maximum temperature above {thresh_tasmax}",
    description="{freq} total length of heat wave events occurring within a given period. "
    "A heat wave occurs when daily minimum and maximum temperatures exceed {thresh_tasmin} and {thresh_tasmax}, "
    "respectively, over at least {window} days.",
    abstract="Total length of heat waves. A heat wave occurs when daily minimum and maximum temperatures exceed "
    "given thresholds for a number of days.",
    cell_methods="",
    keywords="health,",
    compute=indices.heat_wave_total_length,
)


heat_wave_index = Temp(
    title="Heat wave index",
    identifier="heat_wave_index",
    units="days",
    long_name="Total number of days constituting events of at least {window} consecutive days "
    "with daily maximum temperature above {thresh}",
    description="{freq} total number of days that are part of a heatwave within a given period. "
    "A heat wave occurs when daily maximum temperatures exceed {thresh} over at least {window} days.",
    abstract="Number of days that constitute heatwave events. A heat wave occurs when daily minimum and maximum "
    "temperatures exceed given thresholds for a number of days.",
    cell_methods="",
    compute=indices.heat_wave_index,
)

heat_spell_frequency = Temp(
    title="Heat spell frequency",
    identifier="heat_spell_frequency",
    units="",
    long_name="Number of heat spells",
    description="{freq} number of heat spells events. A heat spell occurs when the {window}-day "
    "averages of daily minimum and maximum temperatures each exceed {thresh_tasmin} and {thresh_tasmax}. "
    "All days of the {window}-day period are considered part of the spell. Gaps of fewer than {min_gap} day(s) are allowed "
    "within a spell.",
    abstract="Number of heat spells. A heat spell occurs when rolling averages of daily minimum and maximum temperatures exceed given "
    "thresholds for a number of days.",
    cell_methods="",
    keywords="health,",
    compute=indices.generic.bivariate_spell_length_statistics,
    input={"data1": "tasmin", "data2": "tasmax"},
    parameters=dict(
        spell_reducer="count",
        op=">=",
        window={"default": 3},
        win_reducer={"default": "mean"},
        freq={"default": "YS"},
        threshold1={
            "description": "Threshold for tasmin",
            "default": "20 °C",
            "name": "thresh_tasmin",
        },
        threshold2={
            "description": "Threshold for tasmax",
            "default": "33 °C",
            "name": "thresh_tasmax",
        },
    ),
)

heat_spell_max_length = Temp(
    title="Heat spell maximum length",
    identifier="heat_spell_max_length",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Longest heat spell",
    description="{freq} maximum length of heat spells. A heat spell occurs when the {window}-day "
    "averages of daily minimum and maximum temperatures each exceed {thresh_tasmin} and {thresh_tasmax}. "
    "All days of the {window}-day period are considered part of the spell. Gaps of fewer than {min_gap} day(s) are allowed "
    "within a spell.",
    abstract="The longest heat spell of a period. A heat spell occurs when rolling averages of daily minimum and maximum temperatures exceed given "
    "thresholds for a number of days.",
    compute=indices.generic.bivariate_spell_length_statistics,
    input={"data1": "tasmin", "data2": "tasmax"},
    parameters=dict(
        spell_reducer="max",
        op=">=",
        window={"default": 3},
        win_reducer={"default": "mean"},
        freq={"default": "YS"},
        threshold1={
            "description": "Threshold for tasmin",
            "default": "20 °C",
            "name": "thresh_tasmin",
        },
        threshold2={
            "description": "Threshold for tasmax",
            "default": "33 °C",
            "name": "thresh_tasmax",
        },
    ),
)

heat_spell_total_length = Temp(
    title="Heat spell total length",
    identifier="heat_spell_total_length",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Total length of heat spells.",
    description="{freq} total length of heat spell events. "
    "A heat spell occurs when the {window}-day  averages of daily minimum and maximum temperatures "
    "each exceed {thresh_tasmin} and {thresh_tasmax}.  All days of the {window}-day period are considered part of the spell."
    "Gaps of fewer than {min_gap} day(s) are allowed within a spell.",
    abstract="Total length of heat spells. A heat spell occurs when rolling averages of daily minimum and maximum temperatures exceed given "
    "thresholds for a number of days.",
    compute=indices.generic.bivariate_spell_length_statistics,
    input={"data1": "tasmin", "data2": "tasmax"},
    parameters=dict(
        spell_reducer="sum",
        op=">=",
        window={"default": 3},
        win_reducer={"default": "mean"},
        freq={"default": "YS"},
        threshold1={
            "description": "Threshold for tasmin",
            "default": "20 °C",
            "name": "thresh_tasmin",
        },
        threshold2={
            "description": "Threshold for tasmax",
            "default": "33 °C",
            "name": "thresh_tasmax",
        },
    ),
)

hot_spell_frequency = Temp(
    title="Hot spell frequency",
    identifier="hot_spell_frequency",
    long_name="Number of hot periods of {window} day(s) or more, during which the temperature on a "
    "window of {window} day(s) is above {thresh}.",
    description="The {freq} number of hot periods of {window} day(s) or more, during which the temperature on a "
    "window of {window} day(s) is above {thresh}.",
    abstract="The frequency of hot periods of `N` days or more, during which the temperature "
    "over a given time window of days is above a given threshold.",
    units="",
    cell_methods="",
    compute=indices.hot_spell_frequency,
)

hot_spell_max_length = Temp(
    title="Hot spell maximum length",
    identifier="hot_spell_max_length",
    long_name="Maximum consecutive number of days in a hot period of {window} day(s) or more, during which the "
    "temperature within windows of {window} day(s) is above {thresh}.",
    description="The maximum {freq} number of consecutive days in a hot period of {window} day(s) or more"
    ", during which the temperature within windows of {window} day(s) is above {thresh}.",
    abstract="The maximum length of a hot period of `N` days or more, during which the "
    "temperature over a given time window of days is above a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.hot_spell_max_length,
)

hot_spell_total_length = Temp(
    title="Hot spell total length",
    identifier="hot_spell_total_length",
    long_name="Number of days in hot periods of {window} day(s) or more, during which the"
    "temperature within windows of {window} day(s) is above {thresh}.",
    description="The {freq} number of days in hot periods of {window} day(s) or more, during which the "
    "temperature within windows of {window} day(s) is above {thresh}.",
    abstract="The total length of hot periods of `N` days or more, during which the "
    "temperature over a given time window of days is above a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.hot_spell_total_length,
)

tg_mean = TempWithIndexing(
    title="Mean temperature",
    identifier="tg_mean",
    units="K",
    standard_name="air_temperature",
    long_name="Mean daily mean temperature",
    description="{freq} mean of daily mean temperature.",
    abstract="Mean of daily mean temperature.",
    cell_methods="time: mean over days",
    compute=indices.tg_mean,
)

tg_max = TempWithIndexing(
    title="Maximum of mean temperature",
    identifier="tg_max",
    units="K",
    standard_name="air_temperature",
    long_name="Maximum daily mean temperature",
    description="{freq} maximum of daily mean temperature.",
    abstract="Maximum of daily mean temperature.",
    cell_methods="time: maximum over days",
    compute=indices.tg_max,
)

tg_min = TempWithIndexing(
    title="Minimum of mean temperature",
    identifier="tg_min",
    units="K",
    standard_name="air_temperature",
    long_name="Minimum daily mean temperature",
    description="{freq} minimum of daily mean temperature.",
    abstract="Minimum of daily mean temperature.",
    cell_methods="time: minimum over days",
    compute=indices.tg_min,
)

tx_mean = TempWithIndexing(
    title="Mean of maximum temperature",
    identifier="tx_mean",
    units="K",
    standard_name="air_temperature",
    long_name="Mean daily maximum temperature",
    description="{freq} mean of daily maximum temperature.",
    abstract="Mean of daily maximum temperature.",
    cell_methods="time: mean over days",
    compute=indices.tx_mean,
)

tx_max = TempWithIndexing(
    title="Maximum temperature",
    identifier="tx_max",
    units="K",
    standard_name="air_temperature",
    long_name="Maximum daily maximum temperature",
    description="{freq} maximum of daily maximum temperature.",
    abstract="Maximum of daily maximum temperature.",
    cell_methods="time: maximum over days",
    compute=indices.tx_max,
)

tx_min = TempWithIndexing(
    title="Minimum of maximum temperature",
    identifier="tx_min",
    units="K",
    standard_name="air_temperature",
    long_name="Minimum daily maximum temperature",
    description="{freq} minimum of daily maximum temperature.",
    abstract="Minimum of daily maximum temperature.",
    cell_methods="time: minimum over days",
    compute=indices.tx_min,
)

tn_mean = TempWithIndexing(
    title="Mean of minimum temperature",
    identifier="tn_mean",
    units="K",
    standard_name="air_temperature",
    long_name="Mean daily minimum temperature",
    description="{freq} mean of daily minimum temperature.",
    abstract="Mean of daily minimum temperature.",
    cell_methods="time: mean over days",
    compute=indices.tn_mean,
)

tn_max = TempWithIndexing(
    title="Maximum of minimum temperature",
    identifier="tn_max",
    units="K",
    standard_name="air_temperature",
    long_name="Maximum daily minimum temperature",
    description="{freq} maximum of daily minimum temperature.",
    abstract="Maximum of daily minimum temperature.",
    cell_methods="time: maximum over days",
    compute=indices.tn_max,
)

tn_min = TempWithIndexing(
    title="Minimum temperature",
    identifier="tn_min",
    units="K",
    standard_name="air_temperature",
    long_name="Minimum daily minimum temperature",
    description="{freq} minimum of daily minimum temperature.",
    abstract="Minimum of daily minimum temperature.",
    cell_methods="time: minimum over days",
    compute=indices.tn_min,
)

daily_temperature_range = TempWithIndexing(
    title="Mean of daily temperature range",
    identifier="dtr",
    units="K",
    standard_name="air_temperature",
    long_name="Mean diurnal temperature range",
    description="{freq} mean diurnal temperature range.",
    cell_methods="time range within days time: mean over days",
    abstract="The average difference between the daily maximum and minimum temperatures.",
    compute=indices.daily_temperature_range,
    parameters={"op": "mean"},
)

max_daily_temperature_range = TempWithIndexing(
    title="Maximum of daily temperature range",
    identifier="dtrmax",
    units="K",
    standard_name="air_temperature",
    long_name="Maximum diurnal temperature range",
    description="{freq} maximum diurnal temperature range.",
    cell_methods="time range within days time: max over days",
    abstract="The maximum difference between the daily maximum and minimum temperatures.",
    compute=indices.daily_temperature_range,
    parameters={"op": "max"},
)

daily_temperature_range_variability = TempWithIndexing(
    title="Variability of daily temperature range",
    identifier="dtrvar",
    units="K",
    standard_name="air_temperature",
    long_name="Mean diurnal temperature range variability",
    description="{freq} mean diurnal temperature range variability, defined as the average day-to-day variation "
    "in daily temperature range for the given time period.",
    abstract="The average day-to-day variation in daily temperature range.",
    cell_methods="time range within days time: difference over days time: mean over days",
    compute=indices.daily_temperature_range_variability,
)

extreme_temperature_range = TempWithIndexing(
    title="Extreme temperature range",
    identifier="etr",
    units="K",
    standard_name="air_temperature",
    long_name="Intra-period extreme temperature range",
    description="{freq} range between the maximum of daily maximum temperature and the minimum of daily"
    "minimum temperature.",
    abstract="The maximum of the maximum temperature minus the minimum of the minimum temperature.",
    compute=indices.extreme_temperature_range,
)

cold_spell_duration_index = Temp(
    title="Cold Spell Duration Index (CSDI)",
    identifier="cold_spell_duration_index",
    var_name="csdi_{window}",
    units="days",
    standard_name="cold_spell_duration_index",
    long_name="Total number of days constituting events of at least {window} consecutive days "
    "where the daily minimum temperature is below the {tasmin_per_thresh}th percentile",
    description="{freq} number of days with at least {window} consecutive days where the daily minimum temperature "
    "is below the {tasmin_per_thresh}th percentile. A {tasmin_per_window} day(s) window, centred on each calendar day "
    "in the {tasmin_per_period} period, is used to compute the {tasmin_per_thresh}th percentile(s).",
    abstract="Number of days part of a percentile-defined cold spell. A cold spell occurs when the daily minimum "
    "temperature is below a given percentile for a given number of consecutive days.",
    cell_methods="",
    compute=indices.cold_spell_duration_index,
)

cold_spell_days = Temp(
    title="Cold spell days",
    identifier="cold_spell_days",
    units="days",
    standard_name="cold_spell_days",
    long_name="Total number of days constituting events of at least {window} consecutive days "
    "where the mean daily temperature is below {thresh}",
    description="{freq} number of days that are part of a cold spell. A cold spell is defined as {window} or more "
    "consecutive days with mean daily temperature below {thresh}.",
    abstract="The number of days that are part of a cold spell. A cold spell is defined as a minimum number of "
    "consecutive days with mean daily temperature below a given threshold.",
    cell_methods="",
    compute=indices.cold_spell_days,
)

cold_spell_frequency = Temp(
    title="Cold spell frequency",
    identifier="cold_spell_frequency",
    long_name="Number of cold periods of {window} day(s) or more, during which the temperature on a "
    "window of {window} day(s) is below {thresh}.",
    description="The {freq} number of cold periods of {window} day(s) or more, during which the temperature on a "
    "window of {window} day(s) is below {thresh}.",
    abstract="The frequency of cold periods of `N` days or more, during which the temperature "
    "over a given time window of days is below a given threshold.",
    units="",
    cell_methods="",
    compute=indices.cold_spell_frequency,
)

cold_spell_max_length = Temp(
    title="Cold spell maximum length",
    identifier="cold_spell_max_length",
    long_name="Maximum consecutive number of days in a cold period of {window} day(s) or more, during which the "
    "temperature within windows of {window} day(s) is under {thresh}.",
    description="The maximum {freq} number of consecutive days in a cold period of {window} day(s) or more"
    ", during which the temperature within windows of {window} day(s) is under {thresh}.",
    abstract="The maximum length of a cold period of `N` days or more, during which the "
    "temperature over a given time window of days is below a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.cold_spell_max_length,
)

cold_spell_total_length = Temp(
    title="Cold spell total length",
    identifier="cold_spell_total_length",
    long_name="Number of days in cold periods of {window} day(s) or more, during which the"
    "temperature within windows of {window} day(s) is under {thresh}.",
    description="The {freq} number of days in cold periods of {window} day(s) or more, during which the "
    "temperature within windows of {window} day(s) is under {thresh}.",
    abstract="The total length of cold periods of `N` days or more, during which the "
    "temperature over a given time window of days is below a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.cold_spell_total_length,
)

cool_night_index = Temp(
    title="Cool night index",
    identifier="cool_night_index",
    units="degC",
    long_name="Mean minimum temperature in late summer",
    description="Mean minimum temperature for September (Northern hemisphere) or March (Southern hemisphere).",
    abstract="A night coolness variable which takes into account the mean minimum night temperatures during the "
    "month when ripening usually occurs beyond the ripening period.",
    cell_methods="time: mean over days",
    src_freq=["D", "M"],
    compute=indices.cool_night_index,
)

daily_freezethaw_cycles = TempWithIndexing(
    title="Daily freeze-thaw cycles",
    identifier="dlyfrzthw",
    units="days",
    long_name="Number of days where maximum daily temperatures are above {thresh_tasmax} "
    "and minimum daily temperatures are at or below {thresh_tasmin}",
    description="{freq} number of days with a diurnal freeze-thaw cycle, where maximum daily temperatures are above "
    "{thresh_tasmax} and minimum daily temperatures are at or below {thresh_tasmin}.",
    abstract="The number of days with a freeze-thaw cycle. A freeze-thaw cycle is defined as a day where maximum daily "
    "temperature is above a given threshold and minimum daily temperature is at or below a given threshold, "
    "usually 0°C for both.",
    cell_methods="",
    compute=indices.multiday_temperature_swing,
    parameters={
        "op": "sum",
        "window": 1,
        "thresh_tasmax": {"default": "0 degC"},
        "thresh_tasmin": {"default": "0 degC"},
        "op_tasmax": {"default": ">"},
        "op_tasmin": {"default": "<="},
    },
)


freezethaw_spell_frequency = Temp(
    title="Freeze-thaw spell frequency",
    identifier="freezethaw_spell_frequency",
    units="days",
    long_name="Frequency of events where maximum daily temperatures are above {thresh_tasmax} "
    "and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive day(s).",
    description="{freq} number of freeze-thaw spells, where maximum daily temperatures are above {thresh_tasmax} "
    "and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive day(s).",
    abstract="Frequency of daily freeze-thaw spells. A freeze-thaw spell is defined as a number of consecutive days "
    "where maximum daily temperatures are above a given threshold and minimum daily temperatures are at or below a "
    "given threshold, usually 0°C for both.",
    cell_methods="",
    compute=indices.multiday_temperature_swing,
    parameters={
        "op": "count",
        "thresh_tasmax": {"default": "0 degC"},
        "thresh_tasmin": {"default": "0 degC"},
        "op_tasmax": {"default": ">"},
        "op_tasmin": {"default": "<="},
    },
)


freezethaw_spell_mean_length = Temp(
    title="Freeze-thaw spell mean length",
    identifier="freezethaw_spell_mean_length",
    units="days",
    long_name="Average length of events where maximum daily temperatures are above {thresh_tasmax} "
    "and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive day(s).",
    description="{freq} average length of freeze-thaw spells, where maximum daily temperatures are above "
    "{thresh_tasmax} and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive "
    "day(s).",
    abstract="Average length of daily freeze-thaw spells. A freeze-thaw spell is defined as a number of consecutive "
    "days where maximum daily temperatures are above a given threshold and minimum daily temperatures are at or below "
    "a given threshold, usually 0°C for both.",
    cell_methods="",
    compute=indices.multiday_temperature_swing,
    parameters={
        "op": "mean",
        "thresh_tasmax": {"default": "0 degC"},
        "thresh_tasmin": {"default": "0 degC"},
        "op_tasmax": ">",
        "op_tasmin": "<=",
    },
)


freezethaw_spell_max_length = Temp(
    title="Maximal length of freeze-thaw spells",
    identifier="freezethaw_spell_max_length",
    units="days",
    long_name="Maximal length of events where maximum daily temperatures are above {thresh_tasmax} "
    "and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive day(s).",
    description="{freq} maximal length of freeze-thaw spells, where maximum daily temperatures are above "
    "{thresh_tasmax} and minimum daily temperatures are at or below {thresh_tasmin} for at least {window} consecutive "
    "day(s).",
    abstract="Maximal length of daily freeze-thaw spells. A freeze-thaw spell is defined as a number of consecutive "
    "days where maximum daily temperatures are above a given threshold and minimum daily temperatures are at or below "
    "a threshold, usually 0°C for both.",
    cell_methods="",
    compute=indices.multiday_temperature_swing,
    parameters={
        "op": "max",
        "thresh_tasmax": {"default": "0 degC"},
        "thresh_tasmin": {"default": "0 degC"},
        "op_tasmax": {"default": ">"},
        "op_tasmin": {"default": "<="},
    },
)


cooling_degree_days = TempWithIndexing(
    title="Cooling degree days",
    identifier="cooling_degree_days",
    units="K days",
    standard_name="integral_of_air_temperature_excess_wrt_time",
    long_name="Cumulative sum of temperature degrees for mean daily temperature above {thresh}",
    description="{freq} cumulative cooling degree days (mean temperature above {thresh}).",
    abstract="The cumulative degree days for days when the mean daily temperature is above a given threshold and "
    "buildings must be air conditioned.",
    cell_methods="time: sum over days",
    compute=indices.cooling_degree_days,
    parameters={"thresh": {"default": "18.0 degC"}},
)

heating_degree_days = TempWithIndexing(
    title="Heating degree days",
    identifier="heating_degree_days",
    units="K days",
    standard_name="integral_of_air_temperature_deficit_wrt_time",
    long_name="Cumulative sum of temperature degrees for mean daily temperature below {thresh}",
    description="{freq} cumulative heating degree days (mean temperature below {thresh}).",
    abstract="The cumulative degree days for days when the mean daily temperature is below a given threshold and "
    "buildings must be heated.",
    cell_methods="time: sum over days",
    compute=indices.heating_degree_days,
    parameters={"thresh": {"default": "17.0 degC"}},
)

growing_degree_days = TempWithIndexing(
    title="Growing degree days",
    identifier="growing_degree_days",
    units="K days",
    standard_name="integral_of_air_temperature_excess_wrt_time",
    long_name="Cumulative sum of temperature degrees for mean daily temperature above {thresh}",
    description="{freq} growing degree days (mean temperature above {thresh}).",
    abstract="The cumulative degree days for days when the average temperature is above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.growing_degree_days,
    parameters={"thresh": {"default": "4.0 degC"}},
)

freezing_degree_days = TempWithIndexing(
    title="Freezing degree days",
    identifier="freezing_degree_days",
    units="K days",
    standard_name="integral_of_air_temperature_deficit_wrt_time",
    long_name="Cumulative sum of temperature degrees for mean daily temperature below {thresh}",
    description="{freq} freezing degree days (mean temperature below {thresh}).",
    abstract="The cumulative degree days for days when the average temperature is below a given threshold, "
    "typically 0°C.",
    cell_methods="time: sum over days",
    compute=indices.heating_degree_days,
    parameters={"thresh": {"default": "0 degC"}},
)

thawing_degree_days = TempWithIndexing(
    title="Thawing degree days",
    identifier="thawing_degree_days",
    units="K days",
    standard_name="integral_of_air_temperature_excess_wrt_time",
    long_name="Cumulative sum of temperature degrees for mean daily temperature above {thresh}",
    description="{freq} thawing degree days (mean temperature above {thresh}).",
    abstract="The cumulative degree days for days when the average temperature is above a given threshold, "
    "typically 0°C.",
    cell_methods="time: sum over days",
    compute=indices.growing_degree_days,
    parameters={"thresh": {"default": "0 degC"}},
)

freshet_start = Temp(
    title="Day of year of spring freshet start",
    identifier="freshet_start",
    units="",
    standard_name="day_of_year",
    long_name="First day where temperature threshold of {thresh} is exceeded for at least {window} days",
    description="Day of year of the spring freshet start, defined as the first day a temperature threshold of {thresh} "
    "is exceeded for at least {window} days.",
    abstract="Day of year of the spring freshet start, defined as the first day when the temperature exceeds a certain "
    "threshold for a given number of consecutive days.",
    compute=indices.first_day_temperature_above,
    parameters={"thresh": {"default": "0 degC"}, "window": {"default": 5}},
)

frost_days = Temp(
    title="Frost days",
    identifier="frost_days",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days where the daily minimum temperature is below {thresh}",
    description="{freq} number of days where the daily minimum temperature is below {thresh}.",
    abstract="Number of days where the daily minimum temperature is below a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.frost_days,
)

frost_season_length = Temp(
    title="Frost season length",
    identifier="frost_season_length",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days between the first occurrence of at least {window} consecutive days with "
    "minimum daily temperature below {thresh} and the first occurrence of at least {window} consecutive days with "
    "minimum daily temperature at or above {thresh} after {mid_date}",
    description="{freq} number of days between the first occurrence of at least {window} consecutive days with "
    "minimum daily temperature below {thresh} and the first occurrence of at least {window} consecutive days with "
    "minimum daily temperature at or above {thresh} after {mid_date}.",
    abstract="Duration of the freezing season, defined as the period when the daily minimum temperature is below 0°C "
    "without a thawing window of days, with the thaw occurring after a median calendar date.",
    cell_methods="time: sum over days",
    compute=indices.frost_season_length,
    parameters={"thresh": {"default": "0 degC"}},
)

last_spring_frost = Temp(
    title="Last spring frost",
    identifier="last_spring_frost",
    units="",
    standard_name="day_of_year",
    long_name="Last day of minimum daily temperature below a threshold of {thresh} "
    "for at least {window} days before a given date ({before_date})",
    description="Day of year of last spring frost, defined as the last day a minimum temperature "
    "remains below a threshold of {thresh} for at least {window} days before a given date ({before_date}).",
    abstract="The last day when minimum temperature is below a given threshold for a certain number of days, "
    "limited by a final calendar date.",
    cell_methods="",
    compute=indices.last_spring_frost,
    parameters={"before_date": {"default": "07-01"}},
)

first_day_tn_below = Temp(
    identifier="first_day_tn_below",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of minimum temperature below {thresh}",
    description="First day of year with minimum temperature below {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_below,
    input={"tas": "tasmin"},
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "07-01"},
        "op": {"default": "<"},
    },
)

first_day_tg_below = Temp(
    identifier="first_day_tg_below",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of mean temperature below {thresh}",
    description="First day of year with mean temperature below {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_below,
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "07-01"},
        "op": {"default": "<"},
    },
)

first_day_tx_below = Temp(
    identifier="first_day_tx_below",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of maximum temperature below {thresh}",
    description="First day of year with maximum temperature below {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_below,
    input={"tas": "tasmax"},
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "07-01"},
        "op": {"default": "<"},
    },
)

first_day_tn_above = Temp(
    identifier="first_day_tn_above",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of minimum temperature above {thresh}",
    description="First day of year with minimum temperature above {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_above,
    input={"tas": "tasmin"},
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "01-01"},
        "op": {"default": ">"},
    },
)


first_day_tg_above = Temp(
    identifier="first_day_tg_above",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of mean temperature above {thresh}",
    description="First day of year with mean temperature above {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_above,
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "01-01"},
        "op": {"default": ">"},
    },
)

first_day_tx_above = Temp(
    identifier="first_day_tx_above",
    units="",
    standard_name="day_of_year",
    long_name="First day of year with a period of at least {window} days of maximum temperature above {thresh}",
    description="First day of year with maximum temperature above {thresh} for at least {window} days.",
    compute=indices.first_day_temperature_above,
    input={"tas": "tasmax"},
    parameters={
        "thresh": {"default": "0 degC"},
        "after_date": {"default": "01-01"},
        "op": {"default": ">"},
    },
)

ice_days = TempWithIndexing(
    title="Ice days",
    identifier="ice_days",
    standard_name="days_with_air_temperature_below_threshold",
    units="days",
    long_name="Number of days with maximum daily temperature below {thresh}",
    description="{freq} number of days where the maximum daily temperature is below {thresh}.",
    abstract="Number of days where the daily maximum temperature is below 0°C",
    cell_methods="time: sum over days",
    compute=indices.ice_days,
)

consecutive_frost_days = Temp(
    title="Consecutive frost days",
    identifier="consecutive_frost_days",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_below_threshold",
    long_name="Maximum number of consecutive days where minimum daily temperature is below {thresh}",
    description="{freq} maximum number of consecutive days where minimum daily temperature is below {thresh}.",
    abstract="Maximum number of consecutive days where the daily minimum temperature is below 0°C",
    cell_methods="time: maximum over days",
    compute=indices.maximum_consecutive_frost_days,
    parameters={"thresh": {"default": "0 degC"}},
)

frost_free_season_length = Temp(
    title="Frost free season length",
    identifier="frost_free_season_length",
    units="days",
    standard_name="days_with_air_temperature_above_threshold",
    long_name="Number of days between the first occurrence of at least {window} consecutive days "
    "with minimum daily temperature at or above {thresh} and the first occurrence of at least "
    "{window} consecutive days with minimum daily temperature below {thresh} after {mid_date}",
    description="{freq} number of days between the first occurrence of at least {window} consecutive days "
    "with minimum daily temperature at or above {thresh} and the first occurrence of at least "
    "{window} consecutive days with minimum daily temperature below {thresh} after {mid_date}.",
    abstract="Duration of the frost free season, defined as the period when the minimum daily temperature is above 0°C "
    "without a freezing window of `N` days, with freezing occurring after a median calendar date.",
    cell_methods="time: sum over days",
    compute=indices.frost_free_season_length,
    parameters={"thresh": {"default": "0 degC"}},
)

frost_free_season_start = Temp(
    title="Frost free season start",
    identifier="frost_free_season_start",
    units="",
    standard_name="day_of_year",
    long_name="First day following a period of {window} days with minimum daily temperature at or above {thresh}",
    description="Day of the year of the beginning of the frost-free season, defined as the {window}th consecutive day "
    "when minimum daily temperature exceeds {thresh}.",
    abstract="First day when minimum daily temperature exceeds a given threshold for a given number of consecutive days",
    compute=indices.frost_free_season_start,
    parameters={"thresh": {"default": "0 degC"}},
)

frost_free_season_end = Temp(
    title="Frost free season end",
    identifier="frost_free_season_end",
    units="",
    standard_name="day_of_year",
    long_name="First day, after {mid_date}, following a period of {window} days "
    "with minimum daily temperature below {thresh}",
    description="Day of the year of the end of the frost-free season, defined as the interval between the first set of "
    "{window} days when the minimum daily temperature is at or above {thresh} "
    "and the first set (after {mid_date}) of {window} days when it is below {thresh}.",
    abstract="First day when the temperature is below a given threshold for a given number of consecutive days after "
    "a median calendar date.",
    cell_methods="",
    compute=indices.frost_free_season_end,
    parameters={"thresh": {"default": "0 degC"}},
)

frost_free_spell_max_length = Temp(
    title="Frost free spell maximum length",
    identifier="frost_free_spell_max_length",
    long_name="Maximum consecutive number of days in a frost free period of {window} day(s) or more, during which the "
    "minimum temperature within windows of {window} day(s) is above {thresh}.",
    description="The maximum {freq} number of consecutive days in a frost free period of {window} day(s) or more"
    ", during which the minimum temperature within windows of {window} day(s) is above {thresh}.",
    abstract="The maximum length of a frost free period of `N` days or more, during which the "
    "minimum temperature over a given time window of days is above a given threshold.",
    units="days",
    cell_methods="",
    compute=indices.frost_free_spell_max_length,
)

maximum_consecutive_frost_free_days = Temp(
    title="Maximum consecutive frost free days",
    # FIXME: shouldn't this be `maximum_`? Breaking changes needed.
    identifier="consecutive_frost_free_days",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Maximum number of consecutive days with minimum temperature at or above {thresh}",
    description="{freq} maximum number of consecutive days with minimum daily temperature at or above {thresh}.",
    abstract="Maximum number of consecutive frost-free days: where the daily minimum temperature is above "
    "or equal to 0°C",
    cell_methods="time: maximum over days",
    compute=indices.maximum_consecutive_frost_free_days,
    parameters={"thresh": {"default": "0 degC"}},
)

growing_season_start = Temp(
    title="Growing season start",
    identifier="growing_season_start",
    units="",
    standard_name="day_of_year",
    long_name="First day of the first series of {window} days with mean daily temperature {op} {thresh}",
    description="Day of the year marking the beginning of the growing season, defined as the first day of the first "
    "series of {window} days with mean daily temperature {op} {thresh}.",
    abstract="The first day when the temperature exceeds a certain threshold for a given number of consecutive days.",
    cell_methods="",
    compute=indices.growing_season_start,
    parameters={"thresh": {"default": "5.0 degC"}, "op": {"default": ">="}},
)

growing_season_length = Temp(
    title="Growing season length",
    identifier="growing_season_length",
    units="days",
    standard_name="growing_season_length",
    long_name="Number of days between the first occurrence of at least {window} consecutive days with mean "
    "daily temperature over {thresh} and the first occurrence of at least {window} consecutive days with "
    "mean daily temperature below {thresh}, occurring after {mid_date}",
    description="{freq} number of days between the first occurrence of at least {window} consecutive days "
    "with mean daily temperature over {thresh} and the first occurrence of at least {window} consecutive days with "
    "mean daily temperature below {thresh}, occurring after {mid_date}.",
    abstract="Number of days between the first occurrence of a series of days with a daily average temperature above a "
    "threshold and the first occurrence of a series of days with a daily average temperature below that same "
    "threshold, occurring after a given calendar date.",
    cell_methods="",
    compute=indices.growing_season_length,
    parameters={
        "thresh": {"default": "5.0 degC"},
        "op": {"default": ">="},
        "mid_date": {"default": "07-01"},
    },
)

growing_season_end = Temp(
    title="Growing season end",
    identifier="growing_season_end",
    units="",
    standard_name="day_of_year",
    long_name="First day of the first series of {window} days with mean daily temperature {op} {thresh}, "
    "occurring after {mid_date}",
    description="Day of year of end of growing season, defined as the first day of consistent inferior threshold "
    "temperature of {thresh} after a run of {window} days superior to threshold temperature, occurring after "
    "{mid_date}.",
    abstract="The first day when the temperature is below a certain threshold for a certain number of consecutive days "
    "after a given calendar date.",
    cell_methods="",
    compute=indices.growing_season_end,
    parameters={
        "thresh": {"default": "5.0 degC"},
        "op": {"default": "<"},
        "mid_date": {"default": "07-01"},
    },
)

tropical_nights = TempWithIndexing(
    title="Tropical nights",
    identifier="tropical_nights",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="Number of days with minimum daily temperature above {thresh}",
    description="{freq} number of Tropical Nights, defined as days with minimum daily temperature above {thresh}.",
    abstract="Number of days where minimum temperature is above a given threshold.",
    cell_methods="time: sum over days",
    compute=indices.tn_days_above,
    parameters={"thresh": {"default": "20.0 degC"}},
)

tg90p = TempWithIndexing(
    title="Days with mean temperature above the 90th percentile",
    identifier="tg90p",
    units="days",
    standard_name="days_with_air_temperature_above_threshold",
    long_name="Number of days with mean temperature above the 90th percentile",
    description="{freq} number of days with mean temperature above the 90th percentile. "
    "A {tas_per_window} day(s) window, centered on each calendar day in the {tas_per_period} period, "
    "is used to compute the 90th percentile.",
    abstract="Number of days with mean temperature above the 90th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tg90p,
)

tg10p = TempWithIndexing(
    title="Days with mean temperature below the 10th percentile",
    identifier="tg10p",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days with mean temperature below the 10th percentile",
    description="{freq} number of days with mean temperature below the 10th percentile. "
    "A {tas_per_window} day(s) window, centered on each calendar day in the {tas_per_period} period, "
    "is used to compute the 10th percentile.",
    abstract="Number of days with mean temperature below the 10th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tg10p,
)

tx90p = TempWithIndexing(
    title="Days with maximum temperature above the 90th percentile",
    identifier="tx90p",
    units="days",
    standard_name="days_with_air_temperature_above_threshold",
    long_name="Number of days with maximum temperature above the 90th percentile",
    description="{freq} number of days with maximum temperature above the 90th percentile. "
    "A {tasmax_per_window} day(s) window, centered on each calendar day in the {tasmax_per_period} period, "
    "is used to compute the 90th percentile.",
    abstract="Number of days with maximum temperature above the 90th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tx90p,
)

tx10p = TempWithIndexing(
    title="Days with maximum temperature below the 10th percentile",
    identifier="tx10p",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days with maximum temperature below the 10th percentile",
    description="{freq} number of days with maximum temperature below the 10th percentile. "
    "A {tasmax_per_window} day(s) window, centered on each calendar day in the {tasmax_per_period} period, "
    "is used to compute the 10th percentile.",
    abstract="Number of days with maximum temperature below the 10th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tx10p,
)

tn90p = TempWithIndexing(
    title="Days with minimum temperature above the 90th percentile",
    identifier="tn90p",
    units="days",
    standard_name="days_with_air_temperature_above_threshold",
    long_name="Number of days with minimum temperature above the 90th percentile",
    description="{freq} number of days with minimum temperature above the 90th percentile. "
    "A {tasmin_per_window} day(s) window, centered on each calendar day in the {tasmin_per_period} period, "
    "is used to compute the 90th percentile.",
    abstract="Number of days with minimum temperature above the 90th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tn90p,
)

tn10p = TempWithIndexing(
    title="Days with minimum temperature below the 10th percentile",
    identifier="tn10p",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days with minimum temperature below the 10th percentile",
    description="{freq} number of days with minimum temperature below the 10th percentile. "
    "A {tasmin_per_window} day(s) window, centered on each calendar day in the {tasmin_per_period} period, "
    "is used to compute the 10th percentile.",
    abstract="Number of days with minimum temperature below the 10th percentile.",
    cell_methods="time: sum over days",
    compute=indices.tn10p,
)


degree_days_exceedance_date = Temp(
    title="Degree day exceedance date",
    identifier="degree_days_exceedance_date",
    units="",
    standard_name="day_of_year",
    long_name="Day of year when the integral of mean daily temperature {op} {thresh} exceeds {sum_thresh}",
    description=lambda **kws: "Day of year when the integral of degree days (mean daily temperature {op} {thresh}) "
    "exceeds {sum_thresh}"
    + (
        ", with the cumulative sum starting from {after_date}."
        if kws["after_date"] is not None
        else "."
    ),
    abstract="The day of the year when the sum of degree days exceeds a threshold, occurring after a given date. "
    "Degree days are calculated above or below a given temperature threshold.",
    cell_methods="",
    compute=indices.degree_days_exceedance_date,
)


warm_spell_duration_index = Temp(
    title="Warm spell duration index",
    identifier="warm_spell_duration_index",
    units="days",
    standard_name="number_of_days_with_air_temperature_above_threshold",
    long_name="Number of days with at least {window} consecutive days where the maximum daily temperature is above "
    "the {tasmax_per_thresh}th percentile(s)",
    description="{freq} number of days with at least {window} consecutive days where the maximum daily temperature is "
    "above the {tasmax_per_thresh}th percentile(s). A {tasmax_per_window} day(s) window, centred on each calendar day "
    "in the {tasmax_per_period} period, is used to compute the {tasmax_per_thresh}th percentile(s).",
    abstract="Number of days part of a percentile-defined warm spell. A warm spell occurs when the maximum daily "
    "temperature is above a given percentile for a given number of consecutive days.",
    cell_methods="time: sum over days",
    compute=indices.warm_spell_duration_index,
)


maximum_consecutive_warm_days = Temp(
    title="Maximum consecutive warm days",
    identifier="maximum_consecutive_warm_days",
    units="days",
    standard_name="spell_length_of_days_with_air_temperature_above_threshold",
    long_name="Maximum number of consecutive days with maximum daily temperature above {thresh}",
    description="{freq} longest spell of consecutive days with maximum daily temperature above {thresh}.",
    abstract="Maximum number of consecutive days where the maximum daily temperature exceeds a certain threshold.",
    cell_methods="time: maximum over days",
    compute=indices.maximum_consecutive_tx_days,
)


class FireSeasonBase(Indicator):
    """Special Indicator class for FireSeason that accepts any tas[min/max] and optional snd and is not resampling."""

    keywords = "fire"

    def cfcheck(self, tas, snd=None):
        cfchecks.check_valid(tas, "standard_name", "air_temperature")
        cfchecks.cfcheck_from_name("snd", snd)


fire_season = FireSeasonBase(
    identifier="fire_season",
    description="Fire season mask, computed with method {method}.",
    units="",
    compute=indices.fire_season,
)


huglin_index = Temp(
    title="Huglin heliothermal index",
    identifier="huglin_index",
    units="",
    long_name="Integral of mean daily temperature above {thresh} multiplied by day-length coefficient with {method} "
    "method for days between {start_date} and {end_date}",
    description="Heat-summation index for agroclimatic suitability estimation, developed specifically for viticulture, "
    "computed with {method} formula (Summation of ((Tn + Tx)/2 - {thresh}) * k), where coefficient `k` is a "
    "latitude-based day-length for days between {start_date} and {end_date}.",
    abstract="Heat-summation index for agroclimatic suitability estimation, developed specifically for viticulture. "
    "Considers daily minimum and maximum temperature with a given base threshold, typically between 1 April and 30"
    "September, and integrates a day-length coefficient calculation for higher latitudes. "
    "Metric originally published in Huglin (1978). Day-length coefficient based on Hall & Jones (2010).",
    cell_methods="",
    var_name="hi",
    compute=indices.huglin_index,
    parameters={
        "lat": {"kind": InputKind.VARIABLE},
        "method": {"default": "jones"},
        "start_date": {"default": "04-01"},
        "end_date": {"default": "10-01"},
    },
)


biologically_effective_degree_days = Temp(
    title="Biologically effective degree days",
    identifier="biologically_effective_degree_days",
    units="K days",
    long_name="Integral of mean daily temperature above {thresh_tasmin}, with maximum value of "
    "{max_daily_degree_days}, multiplied by day-length coefficient and temperature range modifier based on {method} "
    "method for days between {start_date} and {end_date}",
    description="Heat-summation index for agroclimatic suitability estimation, developed specifically for viticulture. "
    "Computed with {method} formula (Summation of min((max((Tn + Tx)/2 - {thresh_tasmin}, 0) * k) + TR_adj, Dmax), "
    "where coefficient `k` is a latitude-based day-length for days between {start_date} and {end_date}), "
    "coefficient `TR_adj` is a modifier accounting for large temperature swings, and `Dmax` is the maximum possible"
    "amount of degree days that can be gained within a day ({max_daily_degree_days}).",
    abstract="Considers daily minimum and maximum temperature with a given base threshold between 1 April and 31 "
    "October, with a maximum daily value for cumulative degree days (typically 9°C), and integrates modification "
    "coefficients for latitudes between 40°N and 50°N as well as for swings in daily temperature range. "
    "Metric originally published in Gladstones (1992).",
    cell_methods="",
    var_name="bedd",
    compute=indices.biologically_effective_degree_days,
    parameters={
        "lat": {"kind": InputKind.VARIABLE},
        "method": {"default": "gladstones"},
        "start_date": {"default": "04-01"},
        "end_date": {"default": "11-01"},
    },
)


effective_growing_degree_days = Temp(
    title="Effective growing degree days",
    identifier="effective_growing_degree_days",
    units="K days",
    long_name="Integral of mean daily temperature above {thresh} for days between start and end dates "
    "dynamically determined using {method} method",
    description="Heat-summation index for agroclimatic suitability estimation."
    "Computed with {method} formula (Summation of max((Tn + Tx)/2 - {thresh}, 0) between dynamically-determined "
    "growing season start and end dates. The `bootsma` method uses a 10-day average temperature above {thresh} to "
    "identify a start date, while the `qian` method uses a weighted mean average above {thresh} over 5 days to "
    "determine the start date. The end date of the growing season is the date of first fall frost (Tn < 0°C) occurring"
    "after {after_date}.",
    abstract="Considers daily minimum and maximum temperature with a given base threshold between "
    "dynamically-determined growing season start and end dates. The `bootsma` method uses a 10-day mean temperature "
    "above a given threshold to identify a start date, while the `qian` method uses a weighted mean temperature above "
    "a given threshold over 5 days to determine the start date. The end date of the growing season is the date of "
    "first fall frost (Tn < 0°C) occurring after a given date (typically, July 1). "
    "Metric originally published in Bootsma et al. (2005).",
    cell_methods="",
    var_name="egdd",
    compute=indices.effective_growing_degree_days,
    parameters={
        "method": {"default": "bootsma"},
        "thresh": {"default": "5 degC"},
        "after_date": {"default": "07-01"},
    },
)


latitude_temperature_index = Temp(
    title="Latitude temperature index",
    identifier="latitude_temperature_index",
    units="",
    long_name="Mean temperature of warmest month multiplied by the difference of {lat_factor} minus latitude",
    description="A climate indice based on mean temperature of the warmest month and a latitude-based coefficient to "
    "account for longer day-length favouring growing conditions. Developed specifically for viticulture. "
    "Mean temperature of warmest month multiplied by the difference of {lat_factor} minus latitude.",
    abstract="A climate indice based on mean temperature of the warmest month and a latitude-based coefficient to "
    "account for longer day-length favouring growing conditions. Developed specifically for viticulture. "
    "Mean temperature of warmest month multiplied by the difference of latitude factor coefficient minus latitude. "
    "Metric originally published in Jackson, D. I., & Cherry, N. J. (1988).",
    cell_methods="",
    allowed_periods=["Y"],
    var_name="lti",
    compute=indices.latitude_temperature_index,
    parameters={"lat": {"kind": InputKind.VARIABLE}, "lat_factor": 60},
)


late_frost_days = Temp(
    title="Late frost days",
    identifier="late_frost_days",
    units="days",
    standard_name="days_with_air_temperature_below_threshold",
    long_name="Number of days where the daily minimum temperature is below {thresh}",
    description="{freq} number of days where the daily minimum temperature is below {thresh}"
    "over the period {indexer}.",
    abstract="Number of days where the daily minimum temperature is below a given threshold between a given"
    "start date and a given end date.",
    cell_methods="time: sum over days",
    compute=indices.frost_days,
)


australian_hardiness_zones = Temp(
    title="Australian hardiness zones",
    identifier="australian_hardiness_zones",
    units="",
    long_name="Hardiness zones",
    description="A climate indice based on a {window}-year rolling average of the annual minimum temperature. "
    "Developed specifically to aid in determining plant suitability of geographic regions. The Australian National "
    "Botanical Gardens (ANBG) classification scheme divides categories into 5-degree Celsius zones, starting from -15 "
    "degrees Celsius and ending at 20 degrees Celsius.",
    abstract="A climate indice based on a multi-year rolling average of the annual minimum temperature. "
    "Developed specifically to aid in determining plant suitability of geographic regions. The Australian National "
    "Botanical Gardens (ANBG) classification scheme divides categories into 5-degree Celsius zones, starting from -15 "
    "degrees Celsius and ending at 20 degrees Celsius.",
    cell_methods="",
    allowed_periods=["Y"],
    var_name="hz",
    compute=indices.hardiness_zones,
    parameters={"method": "anbg"},
)


usda_hardiness_zones = Temp(
    title="USDA hardiness zones",
    identifier="usda_hardiness_zones",
    units="",
    long_name="Hardiness zones",
    description="A climate indice based on a {window}-year rolling average of the annual minimum temperature. "
    "Developed specifically to aid in determining plant suitability of geographic regions. The USDA classification"
    "scheme divides categories into 10 degree Fahrenheit zones, with 5-degree Fahrenheit half-zones, "
    "starting from -65 degrees Fahrenheit and ending at 65 degrees Fahrenheit.",
    abstract="A climate indice based on a multi-year rolling average of the annual minimum temperature. "
    "Developed specifically to aid in determining plant suitability of geographic regions. The USDA classification"
    "scheme divides categories into 10 degree Fahrenheit zones, with 5-degree Fahrenheit half-zones, "
    "starting from -65 degrees Fahrenheit and ending at 65 degrees Fahrenheit.",
    cell_methods="",
    allowed_periods=["Y"],
    var_name="hz",
    compute=indices.hardiness_zones,
    parameters={"method": "usda"},
)

chill_portions = TempHourly(
    title="Chill portions",
    identifier="cp",
    units="",
    cell_methods="time: sum",
    description="Chill portions are a measure to estimate the bud breaking potential of different crops. "
    "The constants and functions are taken from Luedeling et al. (2009) which formalises "
    "the method described in Fishman et al. (1987). ",
    abstract="Chill portions are a measure to estimate the bud breaking potential of different crops. "
    "The constants and functions are taken from Luedeling et al. (2009) which formalises "
    "the method described in Fishman et al. (1987). "
    "The model computes the accumulation of cold temperatures in a two-step process. "
    "First, cold temperatures contribute to an intermediate product that is transformed to a chill portion "
    "once it exceeds a certain concentration. The intermediate product can be broken down at higher temperatures "
    "but the final product is stable even at higher temperature. "
    "Thus the dynamic model is more accurate than other chill models like the Chilling hours or Utah model, "
    "especially in moderate climates like Israel, California or Spain.",
    long_name="Chill portions after the Dynamic Model",
    allowed_periods=["Y"],
    compute=indices.chill_portions,
)

chill_units = TempHourlyWithIndexing(
    title="Chill units",
    identifier="cu",
    units="",
    cell_methods="time: sum",
    description="Chill units are a measure to estimate the bud breaking potential of different crops based on the Utah model developed in "
    "Richardson et al. (1974). The Utah model assigns a weight to each hour depending on the temperature recognising that high temperatures can "
    "actually decrease the potential for bud breaking.",
    long_name="Chill units after the Utah Model",
    allowed_periods=["Y"],
    compute=indices.chill_units,
)



# noqa: D205,D400
"""
Uncertainty Partitioning
========================

This module implements methods and tools meant to partition climate projection uncertainties into different components.
"""


from __future__ import annotations

import numpy as np
import pandas as pd
import xarray as xr

# pylint: disable=pointless-string-statement
"""
Implemented partitioning algorithms:

 - `hawkins_sutton`
 - `lafferty_sriver`

# References for other more recent algorithms that could be added here.

Yip, S., Ferro, C. A. T., Stephenson, D. B., and Hawkins, E. (2011). A Simple, Coherent Framework for Partitioning
Uncertainty in Climate Predictions. Journal of Climate 24, 17, 4634-4643, doi:10.1175/2011JCLI4085.1

Northrop, P. J., & Chandler, R. E. (2014). Quantifying sources of uncertainty in projections of future climate.
Journal of Climate, 27(23), 8793–8808, doi:10.1175/JCLI-D-14-00265.1

Goldenson, N., Mauger, G., Leung, L. R., Bitz, C. M., & Rhines, A. (2018). Effects of ensemble configuration on
estimates of regional climate uncertainties. Geophysical Research Letters, 45, 926– 934.
https://doi.org/10.1002/2017GL076297

Lehner, F., Deser, C., Maher, N., Marotzke, J., Fischer, E. M., Brunner, L., Knutti, R., and Hawkins,
E. (2020). Partitioning climate projection uncertainty with multiple large ensembles and CMIP5/6, Earth Syst. Dynam.,
11, 491–508, https://doi.org/10.5194/esd-11-491-2020.

Evin, G., Hingray, B., Blanchet, J., Eckert, N., Morin, S., & Verfaillie, D. (2019). Partitioning Uncertainty
Components of an Incomplete Ensemble of Climate Projections Using Data Augmentation, Journal of Climate, 32(8),
2423-2440, https://doi.org/10.1175/JCLI-D-18-0606.1

Beigi E, Tsai FT-C, Singh VP, Kao S-C. Bayesian Hierarchical Model Uncertainty Quantification for Future Hydroclimate
Projections in Southern Hills-Gulf Region, USA. Water. 2019; 11(2):268. https://doi.org/10.3390/w11020268

Related bixtex entries:
 - yip_2011
 - northrop_2014
 - goldenson_2018
 - lehner_2020
 - evin_2019
"""
# pylint: enable=pointless-string-statement

# TODO: Add ref for Brekke and Barsugli (2013)


def hawkins_sutton(
    da: xr.DataArray,
    sm: xr.DataArray | None = None,
    weights: xr.DataArray | None = None,
    baseline: tuple[str, str] = ("1971", "2000"),
    kind: str = "+",
) -> tuple[xr.DataArray, xr.DataArray]:
    """Return the mean and partitioned variance of an ensemble based on method from Hawkins & Sutton (2009).

    Parameters
    ----------
    da : xr.DataArray
        Time series with dimensions 'time', 'scenario' and 'model'.
    sm : xr.DataArray, optional
        Smoothed time series over time, with the same dimensions as `da`. By default, this is estimated using a
        4th-order polynomial. Results are sensitive to the choice of smoothing function, use this to set another
        polynomial order, or a LOESS curve.
    weights : xr.DataArray, optional
        Weights to be applied to individual models. Should have `model` dimension.
    baseline : (str, str)
        Start and end year of the reference period.
    kind : {'+', '*'}
        Whether the mean over the reference period should be subtracted (+) or divided by (*).

    Returns
    -------
    xr.DataArray, xr.DataArray
        The mean relative to the baseline, and the components of variance of the ensemble. These components are
        coordinates along the `uncertainty` dimension: `variability`, `model`, `scenario`, and `total`.

    Notes
    -----
    To prepare input data, make sure `da` has dimensions `time`, `scenario` and `model`,
    e.g. `da.rename({"scen": "scenario"})`.

    To reproduce results from :cite:t:`hawkins_2009`, input data should meet the following requirements:
      - annual time series starting in 1950 and ending in 2100;
      - the same models are available for all scenarios.

    To get the fraction of the total variance instead of the variance itself, call `fractional_uncertainty` on the
    output.

    References
    ----------
    :cite:cts:`hawkins_2009,hawkins_2011`
    """
    if xr.infer_freq(da.time)[0] not in ["A", "Y"]:
        raise ValueError("This algorithm expects annual time series.")

    if not {"time", "scenario", "model"}.issubset(da.dims):
        raise ValueError(
            "DataArray dimensions should include 'time', 'scenario' and 'model'."
        )

    # Confirm the same models have data for all scenarios
    check = da.notnull().any("time").all("scenario")
    if not check.all():
        raise ValueError(f"Some models are missing data for some scenarios: \n {check}")

    if weights is None:
        weights = xr.ones_like(da.model, float)

    if sm is None:
        # Fit 4th order polynomial to smooth natural fluctuations
        # Note that the order of the polynomial has a substantial influence on the results.
        fit = da.polyfit(dim="time", deg=4, skipna=True)
        sm = xr.polyval(coord=da.time, coeffs=fit.polyfit_coefficients).where(
            da.notnull()
        )

    # Decadal mean residuals
    res = (da - sm).rolling(time=10, center=True).mean()

    # Individual model variance after 2000: V
    # Note that the historical data is the same for all scenarios.
    nv_u = (
        res.sel(time=slice("2000", None))
        .var(dim=("scenario", "time"))
        .weighted(weights)
        .mean("model")
    )

    # Compute baseline average
    ref = sm.sel(time=slice(*baseline)).mean(dim="time")

    # Remove baseline average from smoothed time series
    if kind == "+":
        sm -= ref
    elif kind == "*":
        sm /= ref
    else:
        raise ValueError(kind)

    # Model uncertainty: M(t)
    model_u = sm.weighted(weights).var(dim="model").mean(dim="scenario")

    # Scenario uncertainty: S(t)
    scenario_u = sm.weighted(weights).mean(dim="model").var(dim="scenario")

    # Total uncertainty: T(t)
    total = nv_u + scenario_u + model_u

    # Create output array with the uncertainty components
    u = pd.Index(["variability", "model", "scenario", "total"], name="uncertainty")
    uncertainty = xr.concat([nv_u, model_u, scenario_u, total], dim=u)

    # Keep a trace of the elements for each uncertainty component
    for d in ["model", "scenario"]:
        uncertainty.attrs[d] = da[d].values

    # Mean projection: G(t)
    g = sm.weighted(weights).mean(dim="model").mean(dim="scenario")

    return g, uncertainty


def hawkins_sutton_09_weighting(da, obs, baseline=("1971", "2000")):
    """Return weights according to the ability of models to simulate observed climate change.

    Weights are computed by comparing the 2000 value to the baseline mean: w_m = 1 / (x_{obs} + | x_{m,
    2000} - x_obs | )

    Parameters
    ----------
    da : xr.DataArray
        Input data over the historical period. Should have a time and model dimension.
    obs : float
        Observed change.
    baseline : (str, str)
        Baseline start and end year.

    Returns
    -------
    xr.DataArray
      Weights over the model dimension.
    """
    mm = da.sel(time=slice(*baseline)).mean("time")
    xm = da.sel(time=baseline[1]) - mm
    xm = xm.drop_vars("time").squeeze()
    return 1 / (obs + np.abs(xm - obs))


def lafferty_sriver(
    da: xr.DataArray,
    sm: xr.DataArray | None = None,
    bb13: bool = False,
) -> tuple[xr.DataArray, xr.DataArray]:
    """Return the mean and partitioned variance of an ensemble based on method from Lafferty and Sriver (2023).

    Parameters
    ----------
    da : xr.DataArray
        Time series with dimensions 'time', 'scenario', 'downscaling' and 'model'.
    sm : xr.DataArray
        Smoothed time series over time, with the same dimensions as `da`. By default, this is estimated using a
        4th-order polynomial. Results are sensitive to the choice of smoothing function, use this to set another
        polynomial order, or a LOESS curve.
    bb13 : bool
        Whether to apply the Brekke and Barsugli (2013) method to estimate scenario uncertainty, where the variance
        over scenarios is computed before taking the mean over models and downscaling methods.

    Returns
    -------
    xr.DataArray, xr.DataArray
        The mean relative to the baseline, and the components of variance of the ensemble. These components are
        coordinates along the `uncertainty` dimension: `variability`, `model`, `scenario`, `downscaling` and `total`.

    Notes
    -----
    To prepare input data, make sure `da` has dimensions `time`, `scenario`, `downscaling` and `model`,
    e.g. `da.rename({"experiment": "scenario"})`.

    To get the fraction of the total variance instead of the variance itself, call `fractional_uncertainty` on the
    output.

    References
    ----------
    :cite:cts:`Lafferty2023`
    """
    if xr.infer_freq(da.time)[0] not in ["A", "Y"]:
        raise ValueError("This algorithm expects annual time series.")

    if not {"time", "scenario", "model", "downscaling"}.issubset(da.dims):
        raise ValueError(
            "DataArray dimensions should include 'time', 'scenario', 'downscaling' and 'model'."
        )

    if sm is None:
        # Fit a 4th order polynomial
        fit = da.polyfit(dim="time", deg=4, skipna=True)
        sm = xr.polyval(coord=da.time, coeffs=fit.polyfit_coefficients).where(
            da.notnull()
        )

    # "Interannual variability is then estimated as the centered rolling 11-year variance of the difference
    # between the extracted forced response and the raw outputs, averaged over all outputs."
    nv_u = (
        (da - sm)
        .rolling(time=11, center=True)
        .var()
        .mean(dim=["scenario", "model", "downscaling"])
    )

    # Scenario uncertainty: U_s(t)
    if bb13:
        scenario_u = sm.var(dim="scenario").mean(dim=["model", "downscaling"])
    else:
        scenario_u = sm.mean(dim=["model", "downscaling"]).var(dim="scenario")

    # Model uncertainty: U_m(t)

    # Count the number of parent models that have been downscaled using method $d$ for scenario $s$.
    # In the paper, weights are constant, here they may vary across time if there are missing values.
    mw = sm.count("model")
    # In https://github.com/david0811/lafferty-sriver_2023_npjCliAtm/blob/main/unit_test/lafferty_sriver.py
    # weights are set to zero when there is only one model, but the var for a single element is 0 anyway.
    model_u = sm.var(dim="model").weighted(mw).mean(dim=["scenario", "downscaling"])

    # Downscaling uncertainty: U_d(t)
    dw = sm.count("downscaling")
    downscaling_u = (
        sm.var(dim="downscaling").weighted(dw).mean(dim=["scenario", "model"])
    )

    # Total uncertainty: T(t)
    total = nv_u + scenario_u + model_u + downscaling_u

    # Create output array with the uncertainty components
    u = pd.Index(
        ["model", "scenario", "downscaling", "variability", "total"], name="uncertainty"
    )
    uncertainty = xr.concat([model_u, scenario_u, downscaling_u, nv_u, total], dim=u)

    # Keep a trace of the elements for each uncertainty component
    for d in ["model", "scenario", "downscaling"]:
        uncertainty.attrs[d] = da[d].values

    # Mean projection:
    # This is not part of the original algorithm, but we want all partition algos to have similar outputs.
    g = sm.mean(dim="model").mean(dim="scenario").mean(dim="downscaling")

    return g, uncertainty


def fractional_uncertainty(u: xr.DataArray) -> xr.DataArray:
    """Return the fractional uncertainty.

    Parameters
    ----------
    u : xr.DataArray
        Array with uncertainty components along the `uncertainty` dimension.

    Returns
    -------
    xr.DataArray
        Fractional, or relative uncertainty with respect to the total uncertainty.
    """
    uncertainty = u / u.sel(uncertainty="total") * 100
    uncertainty.attrs.update(u.attrs)
    uncertainty.attrs["long_name"] = "Fraction of total variance"
    uncertainty.attrs["units"] = "%"
    return uncertainty



"""
Data Flags
===========

Pseudo-indicators designed to analyse supplied variables for suspicious/erroneous indicator values.
"""

from __future__ import annotations

from collections.abc import Sequence
from decimal import Decimal
from functools import reduce
from inspect import signature

import numpy as np
import xarray

from xclim.core._exceptions import MissingVariableError, raise_warn_or_log
from xclim.core._types import VARIABLES, Quantified
from xclim.core.calendar import climatological_mean_doy, within_bnds_doy
from xclim.core.formatting import update_xclim_history
from xclim.core.units import convert_units_to, declare_units, infer_context, str2pint
from xclim.core.utils import InputKind, infer_kind_from_parameter
from xclim.indices.generic import binary_ops
from xclim.indices.run_length import suspicious_run

_REGISTRY = {}


class DataQualityException(Exception):
    """Raised when any data evaluation checks are flagged as True.

    Attributes
    ----------
    flag_array : xarray.Dataset
        Xarray.Dataset of Data Flags.
    message : str
        Message prepended to the error messages.
    """

    flag_array: xarray.Dataset | None = None

    def __init__(
        self,
        flag_array: xarray.Dataset,
        message="Data quality flags indicate suspicious values. Flags raised are:\n  - ",
    ):
        self.message = message
        self.flags = []
        for value in flag_array.data_vars.values():
            if value.any():
                for attribute in value.attrs.keys():
                    if str(attribute) == "description":
                        self.flags.append(value.attrs[attribute])
        super().__init__(self.message)

    def __str__(self):
        """Format the errors for history."""
        nl = "\n  - "
        return f"{self.message}{nl.join(self.flags)}"


__all__ = [
    "DataQualityException",
    "data_flags",
    "ecad_compliant",
    "negative_accumulation_values",
    "outside_n_standard_deviations_of_climatology",
    "percentage_values_outside_of_bounds",
    "register_methods",
    "tas_below_tasmin",
    "tas_exceeds_tasmax",
    "tasmax_below_tasmin",
    "temperature_extremely_high",
    "temperature_extremely_low",
    "values_op_thresh_repeating_for_n_or_more_days",
    "values_repeating_for_n_or_more_days",
    "very_large_precipitation_events",
    "wind_values_outside_of_bounds",
]


def register_methods(variable_name=None):
    """Register a data flag functioné.

    Argument can be the output variable name template. The template may use any of the stringable input arguments.
    If not given, the function name is used instead, which may create variable conflicts.
    """

    def _register_methods(func):
        """Summarize all methods used in dataflags checks."""
        func.__dict__["variable_name"] = variable_name or func.__name__
        _REGISTRY[func.__name__] = func
        return func

    return _register_methods


def _sanitize_attrs(da: xarray.DataArray) -> xarray.DataArray:
    to_remove = []
    for attr in da.attrs.keys():
        if str(attr) != "history":
            to_remove.append(attr)
    for attr in to_remove:
        del da.attrs[attr]
    return da


@register_methods()
@update_xclim_history
@declare_units(tasmax="[temperature]", tasmin="[temperature]")
def tasmax_below_tasmin(
    tasmax: xarray.DataArray,
    tasmin: xarray.DataArray,
) -> xarray.DataArray:
    """Check if tasmax values are below tasmin values for any given day.

    Parameters
    ----------
    tasmax : xarray.DataArray
    tasmin : xarray.DataArray

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import tasmax_below_tasmin
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> flagged = tasmax_below_tasmin(ds.tasmax, ds.tasmin)
    """
    tasmax_lt_tasmin = _sanitize_attrs(tasmax < tasmin)
    description = "Maximum temperature values found below minimum temperatures."
    tasmax_lt_tasmin.attrs["description"] = description
    tasmax_lt_tasmin.attrs["units"] = ""
    return tasmax_lt_tasmin


@register_methods()
@update_xclim_history
@declare_units(tas="[temperature]", tasmax="[temperature]")
def tas_exceeds_tasmax(
    tas: xarray.DataArray,
    tasmax: xarray.DataArray,
) -> xarray.DataArray:
    """Check if tas values tasmax values for any given day.

    Parameters
    ----------
    tas : xarray.DataArray
    tasmax : xarray.DataArray

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import tas_exceeds_tasmax
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> flagged = tas_exceeds_tasmax(ds.tas, ds.tasmax)
    """
    tas_gt_tasmax = _sanitize_attrs(tas > tasmax)
    description = "Mean temperature values found above maximum temperatures."
    tas_gt_tasmax.attrs["description"] = description
    tas_gt_tasmax.attrs["units"] = ""
    return tas_gt_tasmax


@register_methods()
@update_xclim_history
@declare_units(tas="[temperature]", tasmin="[temperature]")
def tas_below_tasmin(
    tas: xarray.DataArray, tasmin: xarray.DataArray
) -> xarray.DataArray:
    """Check if tas values are below tasmin values for any given day.

    Parameters
    ----------
    tas : xarray.DataArray
    tasmin : xarray.DataArray

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import tas_below_tasmin
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> flagged = tas_below_tasmin(ds.tas, ds.tasmin)
    """
    tas_lt_tasmin = _sanitize_attrs(tas < tasmin)
    description = "Mean temperature values found below minimum temperatures."
    tas_lt_tasmin.attrs["description"] = description
    tas_lt_tasmin.attrs["units"] = ""
    return tas_lt_tasmin


@register_methods()
@update_xclim_history
@declare_units(da="[temperature]", thresh="[temperature]")
def temperature_extremely_low(
    da: xarray.DataArray, *, thresh: Quantified = "-90 degC"
) -> xarray.DataArray:
    """Check if temperatures values are below -90 degrees Celsius for any given day.

    Parameters
    ----------
    da : xarray.DataArray
    thresh : str

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import temperature_extremely_low
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> temperature = "-90 degC"
    >>> flagged = temperature_extremely_low(ds.tas, thresh=temperature)
    """
    thresh_converted = convert_units_to(thresh, da)
    extreme_low = _sanitize_attrs(da < thresh_converted)
    description = f"Temperatures found below {thresh} in {da.name}."
    extreme_low.attrs["description"] = description
    extreme_low.attrs["units"] = ""
    return extreme_low


@register_methods()
@update_xclim_history
@declare_units(da="[temperature]", thresh="[temperature]")
def temperature_extremely_high(
    da: xarray.DataArray, *, thresh: Quantified = "60 degC"
) -> xarray.DataArray:
    """Check if temperatures values exceed 60 degrees Celsius for any given day.

    Parameters
    ----------
    da : xarray.DataArray
    thresh : str

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import temperature_extremely_high
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> temperature = "60 degC"
    >>> flagged = temperature_extremely_high(ds.tas, thresh=temperature)
    """
    thresh_converted = convert_units_to(thresh, da)
    extreme_high = _sanitize_attrs(da > thresh_converted)
    description = f"Temperatures found in excess of {thresh} in {da.name}."
    extreme_high.attrs["description"] = description
    extreme_high.attrs["units"] = ""
    return extreme_high


@register_methods()
@update_xclim_history
def negative_accumulation_values(
    da: xarray.DataArray,
) -> xarray.DataArray:
    """Check if variable values are negative for any given day.

    Parameters
    ----------
    da : xarray.DataArray

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import negative_accumulation_values
    >>> ds = xr.open_dataset(path_to_pr_file)
    >>> flagged = negative_accumulation_values(ds.pr)
    """
    negative_accumulations = _sanitize_attrs(da < 0)
    description = f"Negative values found for {da.name}."
    negative_accumulations.attrs["description"] = description
    negative_accumulations.attrs["units"] = ""
    return negative_accumulations


@register_methods()
@update_xclim_history
@declare_units(da="[precipitation]", thresh="[precipitation]")
def very_large_precipitation_events(
    da: xarray.DataArray, *, thresh: Quantified = "300 mm d-1"
) -> xarray.DataArray:
    """Check if precipitation values exceed 300 mm/day for any given day.

    Parameters
    ----------
    da : xarray.DataArray
        The DataArray being examined.
    thresh : str
        Threshold to search array for that will trigger flag if any day exceeds value.

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import very_large_precipitation_events
    >>> ds = xr.open_dataset(path_to_pr_file)
    >>> rate = "300 mm d-1"
    >>> flagged = very_large_precipitation_events(ds.pr, thresh=rate)
    """
    thresh_converted = convert_units_to(thresh, da, context="hydro")
    very_large_events = _sanitize_attrs(da > thresh_converted)
    description = f"Precipitation events in excess of {thresh} for {da.name}."
    very_large_events.attrs["description"] = description
    very_large_events.attrs["units"] = ""
    return very_large_events


@register_methods("values_{op}_{thresh}_repeating_for_{n}_or_more_days")
@update_xclim_history
def values_op_thresh_repeating_for_n_or_more_days(
    da: xarray.DataArray, *, n: int, thresh: Quantified, op: str = "=="
) -> xarray.DataArray:
    """Check if array values repeat at a given threshold for `N` or more days.

    Parameters
    ----------
    da : xarray.DataArray
        The DataArray being examined.
    n : int
        Number of days needed to trigger flag.
    thresh : str
        Repeating values to search for that will trigger flag.
    op : {">", "gt", "<", "lt", ">=", "ge", "<=", "le", "==", "eq", "!=", "ne"}
        Operator used for comparison with thresh.

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import values_op_thresh_repeating_for_n_or_more_days
    >>> ds = xr.open_dataset(path_to_pr_file)
    >>> units = "5 mm d-1"
    >>> days = 5
    >>> comparison = "eq"
    >>> flagged = values_op_thresh_repeating_for_n_or_more_days(
    ...     ds.pr, n=days, thresh=units, op=comparison
    ... )
    """
    thresh = convert_units_to(
        thresh, da, context=infer_context(standard_name=da.attrs.get("standard_name"))
    )

    repetitions = _sanitize_attrs(suspicious_run(da, window=n, op=op, thresh=thresh))
    description = (
        f"Repetitive values at {thresh} for at least {n} days found for {da.name}."
    )
    repetitions.attrs["description"] = description
    repetitions.attrs["units"] = ""
    return repetitions


@register_methods()
@update_xclim_history
@declare_units(da="[speed]", lower="[speed]", upper="[speed]")
def wind_values_outside_of_bounds(
    da: xarray.DataArray,
    *,
    lower: Quantified = "0 m s-1",
    upper: Quantified = "46 m s-1",
) -> xarray.DataArray:
    """Check if variable values fall below 0% or rise above 100% for any given day.

    Parameters
    ----------
    da : xarray.DataArray
        The DataArray being examined.
    lower : str
        The lower limit for wind speed.
    upper : str
        The upper limit for wind speed.

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import wind_values_outside_of_bounds
    >>> ceiling, floor = "46 m s-1", "0 m s-1"
    >>> flagged = wind_values_outside_of_bounds(
    ...     sfcWind_dataset, upper=ceiling, lower=floor
    ... )
    """
    lower, upper = convert_units_to(lower, da), convert_units_to(upper, da)
    unbounded_percentages = _sanitize_attrs((da < lower) | (da > upper))
    description = f"Percentage values exceeding bounds of {lower} and {upper} found for {da.name}."
    unbounded_percentages.attrs["description"] = description
    unbounded_percentages.attrs["units"] = ""
    return unbounded_percentages


# TODO: 'Many excessive dry days' = the amount of dry days lies outside a 14·bivariate standard deviation


@register_methods("outside_{n}_standard_deviations_of_climatology")
@update_xclim_history
def outside_n_standard_deviations_of_climatology(
    da: xarray.DataArray,
    *,
    n: int,
    window: int = 5,
) -> xarray.DataArray:
    """Check if any daily value is outside `n` standard deviations from the day of year mean.

    Parameters
    ----------
    da : xarray.DataArray
        The DataArray being examined.
    n : int
        Number of standard deviations.
    window : int
        Moving window used to determining climatological mean. Default: 5.

    Returns
    -------
    xarray.DataArray, [bool]

    Notes
    -----
    A moving window of 5 days is suggested for tas data flag calculations according to ICCLIM data quality standards.

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import outside_n_standard_deviations_of_climatology
    >>> ds = xr.open_dataset(path_to_tas_file)
    >>> std_devs = 5
    >>> average_over = 5
    >>> flagged = outside_n_standard_deviations_of_climatology(
    ...     ds.tas, n=std_devs, window=average_over
    ... )

    References
    ----------
    :cite:cts:`project_team_eca&d_algorithm_2013`
    """
    mu, sig = climatological_mean_doy(da, window=window)
    within_bounds = _sanitize_attrs(
        within_bnds_doy(da, high=(mu + n * sig), low=(mu - n * sig))
    )
    description = (
        f"Values outside of {n} standard deviations from climatology found for {da.name} "
        f"with moving window of {window} days."
    )
    within_bounds.attrs["description"] = description
    within_bounds.attrs["units"] = ""
    return ~within_bounds


@register_methods("values_repeating_for_{n}_or_more_days")
@update_xclim_history
def values_repeating_for_n_or_more_days(
    da: xarray.DataArray, *, n: int
) -> xarray.DataArray:
    """Check if exact values are found to be repeating for at least 5 or more days.

    Parameters
    ----------
    da : xarray.DataArray
        The DataArray being examined.
    n : int
        Number of days to trigger flag.

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import values_repeating_for_n_or_more_days
    >>> ds = xr.open_dataset(path_to_pr_file)
    >>> flagged = values_repeating_for_n_or_more_days(ds.pr, n=5)
    """
    repetition = _sanitize_attrs(suspicious_run(da, window=n))
    description = f"Runs of repetitive values for {n} or more days found for {da.name}."
    repetition.attrs["description"] = description
    repetition.attrs["units"] = ""
    return repetition


@register_methods()
@update_xclim_history
def percentage_values_outside_of_bounds(da: xarray.DataArray) -> xarray.DataArray:
    """Check if variable values fall below 0% or rise above 100% for any given day.

    Parameters
    ----------
    da : xarray.DataArray

    Returns
    -------
    xarray.DataArray, [bool]

    Examples
    --------
    To gain access to the flag_array:

    >>> from xclim.core.dataflags import percentage_values_outside_of_bounds
    >>> flagged = percentage_values_outside_of_bounds(huss_dataset)
    """
    unbounded_percentages = _sanitize_attrs((da < 0) | (da > 100))
    description = f"Percentage values beyond bounds found for {da.name}."
    unbounded_percentages.attrs["description"] = description
    return unbounded_percentages


def data_flags(  # noqa: C901
    da: xarray.DataArray,
    ds: xarray.Dataset | None = None,
    flags: dict | None = None,
    dims: None | str | Sequence[str] = "all",
    freq: str | None = None,
    raise_flags: bool = False,
) -> xarray.Dataset:
    """Evaluate the supplied DataArray for a set of data flag checks.

    Test triggers depend on variable name and availability of extra variables within Dataset for comparison.
    If called with `raise_flags=True`, will raise a DataQualityException with comments for each failed quality check.

    Parameters
    ----------
    da : xarray.DataArray
        The variable to check.
        Must have a name that is a valid CMIP6 variable name and appears in :py:obj:`xclim.core.utils.VARIABLES`.
    ds : xarray.Dataset, optional
        An optional dataset with extra variables needed by some checks.
    flags : dict, optional
        A dictionary where the keys are the name of the flags to check and the values are parameter dictionaries.
        The value can be None if there are no parameters to pass (i.e. default will be used).
        The default, None, means that the data flags list will be taken from :py:obj:`xclim.core.utils.VARIABLES`.
    dims : {"all", None} or str or a sequence of strings
        Dimensions upon which the aggregation should be performed. Default: "all".
    freq : str, optional
        Resampling frequency to have data_flags aggregated over periods.
        Defaults to None, which means the "time" axis is treated as any other dimension (see `dims`).
    raise_flags : bool
        Raise exception if any of the quality assessment flags are raised. Default: False.

    Returns
    -------
    xarray.Dataset

    Examples
    --------
    To evaluate all applicable data flags for a given variable:

    >>> from xclim.core.dataflags import data_flags
    >>> ds = xr.open_dataset(path_to_pr_file)
    >>> flagged = data_flags(ds.pr, ds)
    >>> # The next example evaluates only one data flag, passing specific parameters. It also aggregates the flags
    >>> # yearly over the "time" dimension only, such that a True means there is a bad data point for that year
    >>> # at that location.
    >>> flagged = data_flags(
    ...     ds.pr,
    ...     ds,
    ...     flags={"very_large_precipitation_events": {"thresh": "250 mm d-1"}},
    ...     dims=None,
    ...     freq="YS",
    ... )
    """

    def get_variable_name(function, kwargs):
        fmtargs = {}
        kwargs = kwargs or {}
        for arg, param in signature(function).parameters.items():
            val = kwargs.get(arg, param.default)
            kind = infer_kind_from_parameter(param)
            if arg == "op":
                fmtargs[arg] = val if val not in binary_ops else binary_ops[val]
            elif kind in [
                InputKind.FREQ_STR,
                InputKind.NUMBER,
                InputKind.STRING,
                InputKind.DAY_OF_YEAR,
                InputKind.DATE,
                InputKind.BOOL,
            ]:
                fmtargs[arg] = val
            elif kind == InputKind.QUANTIFIED:
                if isinstance(val, xarray.DataArray):
                    fmtargs[arg] = "array"
                else:
                    val = str2pint(val).magnitude
                    if Decimal(val) % 1 == 0:
                        val = str(int(val))
                    else:
                        val = str(val).replace(".", "point")
                    val = val.replace("-", "minus")
                    fmtargs[arg] = str(val)
        return function.variable_name.format(**fmtargs)

    def _missing_vars(function, dataset: xarray.Dataset, var_provided: str):
        """Handle missing variables in passed datasets."""
        sig = signature(function)
        sig_params = sig.parameters
        extra_vars = {}
        for arg, val in sig_params.items():
            if arg in ["da", var_provided]:
                continue
            kind = infer_kind_from_parameter(val)
            if kind in [InputKind.VARIABLE]:
                if arg in dataset:
                    extra_vars[arg] = dataset[arg]
                else:
                    raise MissingVariableError()
        return extra_vars

    var = str(da.name)
    if dims == "all":
        dims = da.dims
    elif isinstance(dims, str):
        # Thus, a single dimension name, we allow this option to mirror xarray.
        dims = {dims}
    if freq is not None and dims is not None:
        dims = (
            set(dims) - {"time"}
        ) or None  # Will return None if the only dimension was "time".

    if flags is None:
        try:
            flag_funcs = VARIABLES.get(var)["data_flags"]
        except (KeyError, TypeError) as err:
            raise_warn_or_log(
                err,
                mode="raise" if raise_flags else "log",
                msg=f"Data quality checks do not exist for '{var}' variable.",
                err_type=NotImplementedError,
            )
            return xarray.Dataset()
    else:
        flag_funcs = [flags]

    ds = ds or xarray.Dataset()

    flags = {}
    for flag_func in flag_funcs:
        for name, kwargs in flag_func.items():
            func = _REGISTRY[name]
            variable_name = get_variable_name(func, kwargs)
            named_da_variable = None

            try:
                extras = _missing_vars(func, ds, str(da.name))
                # Entries in extras implies that there are two variables being compared
                # Both variables will be sent in as dict entries
                if extras:
                    named_da_variable = {da.name: da}

            except MissingVariableError:
                flags[variable_name] = None
            else:
                with xarray.set_options(keep_attrs=True):
                    if named_da_variable:
                        out = func(**named_da_variable, **extras, **(kwargs or {}))
                    else:
                        out = func(da, **extras, **(kwargs or {}))

                    # Aggregation
                    if freq is not None:
                        out = out.resample(time=freq).any()
                    if dims is not None:
                        out = out.any(dims)

                flags[variable_name] = out

    dsflags = xarray.Dataset(data_vars=flags)

    if raise_flags:
        if np.any([dsflags[dv] for dv in dsflags.data_vars]):
            raise DataQualityException(dsflags)

    return dsflags


def ecad_compliant(
    ds: xarray.Dataset,
    dims: None | str | Sequence[str] = "all",
    raise_flags: bool = False,
    append: bool = True,
) -> xarray.DataArray | xarray.Dataset | None:
    """Run ECAD compliance tests.

    Assert file adheres to ECAD-based quality assurance checks.

    Parameters
    ----------
    ds : xarray.Dataset
        Dataset containing variables to be examined.
    dims : {"all", None} or str or a sequence of strings
        Dimensions upon which aggregation should be performed. Default: ``"all"``.
    raise_flags : bool
        Raise exception if any of the quality assessment flags are raised, otherwise returns None. Default: ``False``.
    append : bool
        If `True`, returns the Dataset with the `ecad_qc_flag` array appended to data_vars.
        If `False`, returns the DataArray of the `ecad_qc_flag` variable.

    Returns
    -------
    xarray.DataArray or xarray.Dataset or None
    """
    flags = xarray.Dataset()
    history: list[str] = []
    for var in ds.data_vars:
        df = data_flags(ds[var], ds, dims=dims)
        for flag_name, flag_data in df.data_vars.items():
            flags = flags.assign({f"{var}_{flag_name}": flag_data})

            if (
                "history" in flag_data.attrs.keys()
                and np.all(flag_data.values) is not None
            ):
                # The extra `split("\n") should be removed when merge_attributes(missing_str=None)
                history_elems = flag_data.attrs["history"].split("\n")[-1].split(" ")
                if not history:
                    history.append(
                        " ".join(
                            [
                                " ".join(history_elems[0:2]),
                                " ".join(history_elems[-4:]),
                                "- Performed the following checks:",
                            ]
                        )
                    )
                history.append(" ".join(history_elems[3:-4]))

    if raise_flags:
        if np.any([flags[dv] for dv in flags.data_vars]):
            raise DataQualityException(flags)
        return None

    ecad_flag = xarray.DataArray(
        # TODO: Test for this change concerning data of type None in dataflag variables
        ~reduce(
            np.logical_or,
            filter(lambda x: x.dtype == bool, flags.data_vars.values()),  # noqa
        ),
        name="ecad_qc_flag",
        attrs={
            "comment": "Adheres to ECAD quality control checks.",
            "history": "\n".join(history),
        },
    )

    if append:
        return xarray.merge([ds, ecad_flag])
    return ecad_flag



"""
Atmospheric Indicators
======================

While the `indices` module stores the computing functions, this module defines Indicator classes and instances that
include a number of functionalities, such as input validation, unit conversion, output meta-data handling,
and missing value masking.

The concept followed here is to define Indicator subclasses for each input variable, then create instances
for each indicator.

"""

from __future__ import annotations

from ._conversion import *
from ._precip import *
from ._synoptic import *
from ._temperature import *
from ._wind import *



from __future__ import annotations

import numpy as np
import pytest

from xclim.core.calendar import percentile_doy
from xclim.indices import (
    cold_spell_duration_index,
    days_over_precip_thresh,
    fraction_over_precip_thresh,
    tg10p,
    tg90p,
    tn10p,
    tn90p,
    tx10p,
    tx90p,
    warm_spell_duration_index,
)
from xclim.testing.helpers import test_timeseries as _test_timeseries


class Test_bootstrap:
    @pytest.mark.slow
    @pytest.mark.parametrize("use_dask", [True, False])
    @pytest.mark.parametrize(
        "var,p,index,freq, cftime",
        (
            ["tas", 98, tg90p, "MS", False],
            ["tasmin", 98, tn90p, "YS-JUL", False],
            ["tasmax", 98, tx90p, "QS-APR", False],
            ["tasmax", 98, tx90p, "QS-APR", True],
            ["tasmin", 2, tn10p, "MS", False],
            ["tasmax", 2, tx10p, "YS", False],
            ["tasmax", 2, tx10p, "YS", True],
            ["tas", 2, tg10p, "MS", False],
            ["tasmax", 98, warm_spell_duration_index, "MS", False],
            ["tasmin", 2, cold_spell_duration_index, "MS", False],
            ["pr", 99, days_over_precip_thresh, "MS", False],
            ["pr", 98, fraction_over_precip_thresh, "MS", False],
            ["pr", 98, fraction_over_precip_thresh, "MS", True],
        ),
    )
    def test_bootstrap(self, var, p, index, freq, cftime, use_dask, random):
        # -- GIVEN
        arr = self.ar1(
            alpha=0.8, n=int(4 * 365.25), random=random, positive_values=(var == "pr")
        )
        climate_var = _test_timeseries(
            arr, start="2000-01-01", variable=var, cftime=cftime
        )
        if use_dask:
            climate_var = climate_var.chunk(dict(time=50))
        in_base_slice = slice("2000-01-01", "2001-12-31")
        out_base_slice = slice("2002-01-01", "2003-12-31")
        per = percentile_doy(climate_var.sel(time=in_base_slice), per=p)

        # -- WHEN
        no_bootstrap = index(climate_var, per, freq=freq, bootstrap=False)
        no_bs_in_base = no_bootstrap.sel(time=in_base_slice)
        no_bs_out_base = no_bootstrap.sel(time=out_base_slice)
        bootstrap = index(climate_var, per, freq=freq, bootstrap=True)
        bootstrapped_in_base = bootstrap.sel(time=in_base_slice)
        bs_out_base = bootstrap.sel(time=out_base_slice)

        # -- THEN
        # Bootstrapping should increase the computed index values within the overlapping
        # period. However, this will not work on unrealistic values such as a constant
        # temperature.
        # Beside, bootstrapping is particularly effective on extreme percentiles, but
        # the closer the target percentile is to the median the less bootstrapping is
        # necessary.
        # Following assertions may even fail if chosen percentile is close to 50.
        assert np.count_nonzero(
            bootstrapped_in_base > no_bs_in_base
        ) > np.count_nonzero(bootstrapped_in_base < no_bs_in_base)
        # bootstrapping should leave the out of base unchanged,
        # but precision above 15th decimal might differ.
        np.testing.assert_array_almost_equal(no_bs_out_base, bs_out_base, 15)

    @pytest.mark.slow
    def test_bootstrap_fraction_over_precip_error_no_doy(self, pr_series):
        with pytest.raises(KeyError):
            # no "dayofyear" coords on per
            fraction_over_precip_thresh(
                pr_series([1, 2]), pr_series([1, 2]), bootstrap=True
            )

    def test_bootstrap_days_over_precip_thresh_error_no_doy(self, pr_series):
        with pytest.raises(KeyError):
            # no "dayofyear" coords on per
            days_over_precip_thresh(
                pr_series([1, 2]), pr_series([1, 2]), bootstrap=True
            )

    def test_bootstrap_no_doy(self, tas_series):
        # no "dayofyear" coords on per
        with pytest.raises(KeyError):
            tg10p(tas_series([42]), tas_series([42]), freq="MS", bootstrap=True)

    def test_bootstrap_full_overlap(self, tas_series, random):
        # bootstrap is unnecessary when in base and out of base fully overlap
        # -- GIVEN
        tas = tas_series(
            self.ar1(alpha=0.8, n=int(4 * 365.25), random=random), start="2000-01-01"
        )
        per = percentile_doy(tas, per=90)
        # -- THEN
        with pytest.raises(KeyError):
            # -- WHEN
            tg10p(tas, per, freq="YS", bootstrap=True)

    @pytest.mark.slow
    def test_bootstrap_no_overlap(self, tas_series, random):
        # bootstrap is unnecessary when in base and out of base fully overlap
        # -- GIVEN
        tas = tas_series(
            self.ar1(alpha=0.8, n=int(4 * 365.25), random=random), start="2000-01-01"
        )
        tas_in_base = tas.sel(time=slice("2000-01-01", "2001-12-31"))
        tas_out_base = tas.sel(time=slice("2002-01-01", "2001-12-31"))
        per = percentile_doy(tas_in_base, per=90)
        # -- THEN
        with pytest.raises(KeyError):
            # -- WHEN
            tg10p(tas_out_base, per, freq="MS", bootstrap=True)

    @pytest.mark.slow
    def test_multi_per(self, open_dataset):
        tas = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").tas
        t90 = percentile_doy(
            tas.sel(time=slice("1990-01-01", "1991-12-31")), window=5, per=[90, 91]
        )
        res = tg90p(tas=tas, tas_per=t90, freq="YS", bootstrap=True)
        np.testing.assert_array_equal([90, 91], res.percentiles)

    @pytest.mark.slow
    def test_doctest_ndims(self, open_dataset):
        """Replicates doctest to facilitate debugging."""
        tas = open_dataset("ERA5/daily_surface_cancities_1990-1993.nc").tas
        t90 = percentile_doy(
            tas.sel(time=slice("1990-01-01", "1991-12-31")), window=5, per=90
        )
        tg90p(tas=tas, tas_per=t90.isel(percentiles=0), freq="YS", bootstrap=True)
        tg90p(tas=tas, tas_per=t90, freq="YS", bootstrap=True)

    def ar1(self, alpha, n, random, positive_values=False):
        """Return "random" AR1 DataArray."""
        # White noise
        wn = random.standard_normal(n - 1) * np.sqrt(1 - alpha**2)
        # Autoregressive series of order 1
        out = np.empty(n)
        out[0] = random.standard_normal()
        for i, w in enumerate(wn):
            if positive_values:
                out[i + 1] = np.abs(alpha * out[i] + w)
            else:
                out[i + 1] = alpha * out[i] + w
        return out



from __future__ import annotations

import numpy as np
import pandas as pd
import pint
import pint.errors
import pytest
import xarray as xr
from dask import array as dsk

from xclim import indices, set_options
from xclim.core import Quantified, ValidationError
from xclim.core.units import (
    amount2lwethickness,
    amount2rate,
    check_units,
    convert_units_to,
    declare_relative_units,
    declare_units,
    infer_context,
    lwethickness2amount,
    pint2cfattrs,
    pint2cfunits,
    pint_multiply,
    rate2amount,
    str2pint,
    to_agg_units,
    units,
    units2pint,
)


class TestUnits:
    def test_temperature(self):
        assert 4 * units.d == 4 * units.day
        Q_ = units.Quantity  # noqa
        assert Q_(1, units.C) == Q_(1, units.degC)

    def test_hydro(self):
        with pytest.raises(pint.errors.DimensionalityError):
            convert_units_to("1 kg m-2", "m")

        with units.context("hydro"):
            q = 1 * units.kg / units.m**2 / units.s
            assert q.to("mm/day") == q.to("mm/d")
            assert q.to("mmday").magnitude == 24 * 60**2

    def test_lat_lon(self):
        assert 100 * units.degreeN == 100 * units.degree

    def test_pcic(self):
        with units.context("hydro"):
            fu = units.parse_units("kilogram / d / meter ** 2")
            tu = units.parse_units("mm/day")
            np.isclose(1 * fu, 1 * tu)

    def test_dimensionality(self):
        with units.context("hydro"):
            fu = 1 * units.parse_units("kg / m**2 / s")
            tu = 1 * units.parse_units("mm / d")
            fu.to("mmday")
            tu.to("mmday")

    def test_fraction(self):
        q = 5 * units.percent
        assert q.to("dimensionless") == 0.05

        q = 5 * units.parse_units("pct")
        assert q.to("dimensionless") == 0.05


class TestConvertUnitsTo:
    def test_deprecation(self, tas_series):
        with pytest.raises(TypeError):
            convert_units_to(0, units.K)

        with pytest.raises(TypeError):
            convert_units_to(10.0, units.mm / units.day, context="hydro")

        with pytest.raises(TypeError):
            tas = tas_series(np.arange(365), start="1/1/2001")
            out = indices.tx_days_above(tas, 30)  # noqa

    def test_fraction(self):
        out = convert_units_to(xr.DataArray([10], attrs={"units": "%"}), "")
        assert out == 0.1

    def test_lazy(self, pr_series):
        pr = pr_series(np.arange(365), start="1/1/2001").chunk({"time": 100})
        out = convert_units_to(pr, "mm/day", context="hydro")
        assert isinstance(out.data, dsk.Array)

    @pytest.mark.parametrize(
        "alias", [units("Celsius"), units("degC"), units("C"), units("deg_C")]
    )
    def test_temperature_aliases(self, alias):
        assert alias == units("celsius")

    def test_offset_confusion(self):
        out = convert_units_to("10 degC days", "K days")
        assert out == 10

    def test_cf_conversion_amount2lwethickness_error(self):
        # It is not thickness data because the standard name is wrong (absent)
        not_thickness_data = xr.DataArray([1, 2, 3], attrs={"units": "mm"})
        with pytest.raises(pint.errors.DimensionalityError):
            convert_units_to(not_thickness_data, "kg/m**2/day")

    def test_cf_conversion_amount2lwethickness_amount2rate(self):
        thickness_data = xr.DataArray(
            [1, 2, 3],
            coords={"time": pd.date_range("1990-01-01", periods=3, freq="D")},
            dims=["time"],
            attrs={"units": "mm", "standard_name": "thickness_of_rainfall_amount"},
        )
        out = convert_units_to(thickness_data, "kg/m**2/day")
        np.testing.assert_array_almost_equal(out, thickness_data)
        assert out.attrs["units"] == "kg d-1 m-2"  # CF equivalent unit
        assert out.attrs["standard_name"] == "rainfall_flux"

    def test_temperature_difference(self):
        delta = xr.DataArray(
            [2], attrs={"units": "K", "units_metadata": "temperature: difference"}
        )
        out = convert_units_to(source=delta, target="delta_degC")
        assert out == 2
        assert out.attrs["units"] == "degC"


class TestUnitConversion:
    def test_pint2cfunits(self):
        u = units("mm/d")
        assert pint2cfunits(u.units) == "mm d-1"

        u = units("percent")
        assert pint2cfunits(u.units) == "%"

        u = units("pct")
        assert pint2cfunits(u.units) == "%"

    def test_units2pint(self, pr_series):
        u = units2pint(pr_series([1, 2]))
        assert pint2cfunits(u) == "kg m-2 s-1"

        u = units2pint("m^3 s-1")
        assert pint2cfunits(u) == "m3 s-1"

        u = units2pint("%")
        assert pint2cfunits(u) == "%"

        u = units2pint("1")
        assert pint2cfunits(u) == "1"

    def test_pint_multiply(self, pr_series):
        a = pr_series([1, 2, 3])
        out = pint_multiply(a, 1 * units.days)
        assert out[0] == 1 * 60 * 60 * 24
        assert out.units == "kg m-2"

    def test_str2pint(self):
        Q_ = units.Quantity  # noqa
        assert str2pint("-0.78 m") == Q_(-0.78, units="meter")
        assert str2pint("m kg/s") == Q_(1, units="meter kilogram/second")
        assert str2pint("11.8 degC days") == Q_(11.8, units="delta_degree_Celsius days")
        assert str2pint("nan m^2 K^-3").units == Q_(1, units="m²/K³").units


class TestCheckUnits:
    def test_basic(self):
        check_units("%", "[]")
        check_units("pct", "[]")
        check_units("mm/day", "[precipitation]")
        check_units("mm/s", "[precipitation]")
        check_units("kg/m2/s", "[precipitation]")
        check_units("cms", "[discharge]")
        check_units("m3/s", "[discharge]")
        check_units("m/s", "[speed]")
        check_units("km/h", "[speed]")
        check_units("degC", "[temperature]")

        with units.context("hydro"):
            check_units("kg/m2", "[length]")

        with set_options(data_validation="raise"):
            with pytest.raises(ValidationError):
                check_units("mm", "[precipitation]")

            with pytest.raises(ValidationError):
                check_units("m3", "[discharge]")

    def test_comparison(self):
        """Check that both units have the same dimensions."""
        check_units("mm/day", "m/hour")

        with pytest.raises(ValidationError):
            check_units("mm/day", "m")

        check_units(
            xr.DataArray([1], attrs={"units": "degC"}),
            xr.DataArray([1], attrs={"units": "degK"}),
        )

        with pytest.raises(ValidationError):
            check_units(xr.DataArray([1], attrs={"units": "degC"}), "2 mm")

        with pytest.raises(ValidationError):
            """There is no context information to know that mm/day is a precipitation unit."""
            check_units("kg/m2/s", "mm/day")

    def test_user_error(self):
        with pytest.raises(ValidationError):
            check_units("deg C", "[temperature]")


def test_rate2amount(pr_series):
    pr = pr_series(np.ones(365 + 366 + 365), start="2019-01-01")

    am_d = rate2amount(pr)
    np.testing.assert_array_equal(am_d, 86400)

    with xr.set_options(keep_attrs=True):
        pr_ms = pr.resample(time="MS").mean()
        pr_m = pr.resample(time="ME").mean()

        am_ms = rate2amount(pr_ms)
        np.testing.assert_array_equal(am_ms[:4], 86400 * np.array([31, 28, 31, 30]))
        am_m = rate2amount(pr_m)
        np.testing.assert_array_equal(am_m[:4], 86400 * np.array([31, 28, 31, 30]))
        np.testing.assert_array_equal(am_ms, am_m)

        pr_ys = pr.resample(time="YS").mean()
        am_ys = rate2amount(pr_ys)

        np.testing.assert_array_equal(am_ys, 86400 * np.array([365, 366, 365]))


@pytest.mark.parametrize(
    "srcfreq, exp", [("h", 3600), ("min", 60), ("s", 1), ("ns", 1e-9)]
)
def test_rate2amount_subdaily(srcfreq, exp):
    pr = xr.DataArray(
        np.ones(1000),
        dims=("time",),
        coords={"time": xr.date_range("2019-01-01", periods=1000, freq=srcfreq)},
        attrs={"units": "kg m-2 s-1"},
    )
    am = rate2amount(pr)
    np.testing.assert_array_equal(am, exp)


def test_amount2rate(pr_series):
    pr = pr_series(np.ones(365 + 366 + 365), start="2019-01-01")
    am = rate2amount(pr)
    assert am.attrs["standard_name"] == "precipitation_amount"

    np.testing.assert_allclose(amount2rate(am), pr)

    with xr.set_options(keep_attrs=True):
        am_ms = am.resample(time="MS").sum()
        am_m = am.resample(time="ME").sum()

        pr_ms = amount2rate(am_ms)
        np.testing.assert_allclose(pr_ms, 1)
        pr_m = amount2rate(am_m)
        np.testing.assert_allclose(pr_m, 1)

        am_ys = am.resample(time="YS").sum()
        pr_ys = amount2rate(am_ys)
        np.testing.assert_allclose(pr_ys, 1)


def test_amount2lwethickness(snw_series):
    snw = snw_series(np.ones(365), start="2019-01-01")

    swe = amount2lwethickness(snw, out_units="mm")
    assert swe.attrs["standard_name"] == "lwe_thickness_of_surface_snow_amount"
    np.testing.assert_allclose(swe, 1)

    snw = lwethickness2amount(swe)
    assert snw.attrs["standard_name"] == "surface_snow_amount"


@pytest.mark.parametrize(
    "std_name,dim,exp",
    [
        ("precipitation_flux", None, "hydro"),
        ("snowfall_flux", None, "none"),
        ("air_temperature", "[precipitation]", "hydro"),
        (None, None, "none"),
    ],
)
def test_infer_context(std_name, dim, exp):
    assert infer_context(std_name, dim) == exp


def test_declare_units():
    """Test that an error is raised when parameters with type Quantified do not declare their dimensions.

    In this example, `wo` is a Quantified parameter, but does not declare its dimension as [length].
    """
    with pytest.raises(ValueError):

        @declare_units(pr="[precipitation]", evspsblpot="[precipitation]")
        def dryness_index(
            pr: xr.DataArray,
            evspsblpot: xr.DataArray,
            lat: xr.DataArray | str | None = None,
            wo: Quantified = "200 mm",
            freq: str = "YS",
        ) -> xr.DataArray:
            pass


def test_declare_relative_units():
    def index(
        data: xr.DataArray, thresh: Quantified, dthreshdt: Quantified  # noqa: F841
    ):
        return xr.DataArray(1, attrs={"units": "rad"})

    index_relative = declare_relative_units(thresh="<data>", dthreshdt="<data>/[time]")(
        index
    )
    assert hasattr(index_relative, "relative_units")

    index_full_mm = declare_units(data="mm")(index_relative)
    assert index_full_mm.in_units == {
        "data": "mm",
        "thresh": "(mm)",
        "dthreshdt": "(mm)/[time]",
    }

    index_full_area = declare_units(data="[area]")(index_relative)
    assert index_full_area.in_units == {
        "data": "[area]",
        "thresh": "([area])",
        "dthreshdt": "([area])/[time]",
    }

    # No failures
    index_full_mm("1 mm", "2 km", "3 mm/s")

    with pytest.raises(ValidationError):
        index_full_mm("1 mm", "2 Pa", "3 mm/s")


@pytest.mark.parametrize(
    "in_u,opfunc,op,exp,exp_u",
    [
        ("m/h", "sum", "integral", 8760, "m"),
        ("m/h", "sum", "sum", 365, "m/h"),
        ("K", "mean", "mean", 1, "K"),
        ("", "sum", "count", 365, "d"),
        ("", "sum", "count", 365, "d"),
        ("kg m-2", "var", "var", 0, "kg2 m-4"),
        (
            "°C",
            "argmax",
            "doymax",
            0,
            "1",
        ),
        (
            "°C",
            "sum",
            "integral",
            365,
            ("degC d", "d degC"),
        ),  # dependent on numpy/pint version
        ("°F", "sum", "integral", 365, "d degF"),  # not sure why the order is different
    ],
)
def test_to_agg_units(in_u, opfunc, op, exp, exp_u):
    da = xr.DataArray(
        np.ones((365,)),
        dims=("time",),
        coords={"time": xr.cftime_range("1993-01-01", periods=365, freq="D")},
        attrs={"units": in_u},
    )
    if units(in_u).dimensionality == "[temperature]":
        da.attrs["units_metadata"] = "temperature: difference"

    # FIXME: This is emitting warnings from deprecated DataArray.argmax() usage.
    out = to_agg_units(getattr(da, opfunc)(), da, op)
    np.testing.assert_allclose(out, exp)
    if isinstance(exp_u, tuple):
        assert out.attrs["units"] in exp_u
    else:
        assert out.attrs["units"] == exp_u


def test_pint2cfattrs():
    attrs = pint2cfattrs(units.degK, is_difference=True)
    assert attrs == {"units": "K", "units_metadata": "temperature: difference"}

    attrs = pint2cfattrs(units.meter, is_difference=True)
    assert "units_metadata" not in attrs

    attrs = pint2cfattrs(units.delta_degC)
    assert attrs == {"units": "degC", "units_metadata": "temperature: difference"}


def test_temp_difference_rountrip():
    """Test roundtrip of temperature difference units."""
    attrs = {"units": "degC", "units_metadata": "temperature: difference"}
    da = xr.DataArray([1], attrs=attrs)
    pu = units2pint(da)
    # Confirm that we get delta pint units
    assert pu == units.delta_degC

    # and that converting those back to cf attrs gives the same result
    attrs = pint2cfattrs(pu)
    assert attrs == {"units": "degC", "units_metadata": "temperature: difference"}
